#
# Copyright 2017-2018 Ettus Research, a National Instruments Company
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""UHD Python API module."""

import os

# As of python3.8, python stopped using PATH to search for DLLs on Windows, and
# required the use of os.add_dll_directory to add paths to the DLL search.
if os.name == "nt":
    uhd_dll_name = "uhd.dll"
    uhd_dll_dir = None
    if hasattr(os, "add_dll_directory"):
        # First, check for a user defined path to the UHD libraries.
        # This might be the case for files built from source but not installed yet.
        uhd_dll_dir = os.environ.get("UHD_LIBRARY_PATH")
        if not uhd_dll_dir or not os.path.exists(os.path.join(uhd_dll_dir, uhd_dll_name)):
            # The UHD_PKG_PATH will be set if UHD was installed with a UHD NSIS installer,
            # use this path to find the uhd.dll and add it to the DLL search.
            uhd_pkg_dir = os.environ.get("UHD_PKG_PATH")
            if uhd_pkg_dir is not None and os.path.exists(
                os.path.join(uhd_pkg_dir, "bin", uhd_dll_name)
            ):
                uhd_dll_dir = os.path.join(uhd_pkg_dir, "bin")
            else:
                # If UHD_PKG_PATH is not set, the UHD may have been built and installed from source.
                # CMake will generate the UHD_INSTALL_PATH variable in the config_vars.py file using the
                # CMAKE_INSTALL_PREFIX, try to find the uhd.dll with this path.
                UHD_INSTALL_PATH = "/usr"
                if os.path.exists(os.path.join(UHD_INSTALL_PATH, "bin", uhd_dll_name)):
                    uhd_dll_dir = os.path.join(UHD_INSTALL_PATH, "bin")
        # perform version check if uhd.dll path was found
        if uhd_dll_dir:
            # compare version of uhd.dll against the version used for the cmake build process
            # potentially this could be a different version, if the NSIS installer did perform
            # an upgrade of uhd.dll separately
            UHD_VERSION = "4.9.0.makepkg-0-g9ec1f582"
            from .utils.version_check import compare_versions, get_windll_version

            os.add_dll_directory(uhd_dll_dir)
            uhd_dll_version = get_windll_version(os.path.join(uhd_dll_dir, uhd_dll_name))
            if compare_versions(uhd_dll_version, UHD_VERSION) != 0:
                print(
                    f"WARNING: Version conflict between {uhd_dll_name}({uhd_dll_version}) and "
                    f"the Python API build version({UHD_VERSION})!\nIt is recommended to align "
                    "the UHD binary installation and the UHD Python module to the exact same version "
                    "by upgrading the older component."
                )
    if not uhd_dll_dir:
        print(
            f"WARNING: Failed to add {uhd_dll_name} path to DLL search, module may not import properly, "
            "version check skipped.\nPlease ensure UHD_PKG_DIR (binary installation) is set correctly. "
            'For Python version < 3.8 the "UHD\\bin" directory needs to be added to PATH environment variable.'
        )

try:
    from . import chdr, dsp, filters, rfnoc, types, usrp, usrp_clock, usrpctl
    from .libpyuhd import find, get_abi_string, get_component, get_version_string
    from .libpyuhd.paths import *  # noqa: F403
    from .property_tree import PropertyTree

    __version__ = get_version_string()
except ImportError as e:
    enable_pymod_utils = "ON"
    if enable_pymod_utils != "ON":
        raise e

    def get_tmp_path():
        """Return the path to the temporary directory (Python-only version)."""
        env_path = os.environ.get("UHD_TEMP_PATH")
        if env_path is not None and os.path.isdir(env_path):
            return env_path
        import tempfile

        return tempfile.gettempdir()

    def get_pkg_path():
        """Return the path to the UHD package directory (Python-only version)."""
        env_path = os.environ.get("UHD_PKG_PATH")
        if env_path is not None and os.path.isdir(env_path):
            return env_path
        return "/usr"

    def get_pkg_data_path():
        """Return the path to the UHD package directory (Python-only version)."""
        env_path = os.environ.get("UHD_PKG_DATA_PATH")
        if env_path is not None and os.path.isdir(env_path):
            return env_path
        return os.path.join("/usr", "share/uhd")

    __version__ = "4.9.0.makepkg-0-g9ec1f582"



#
# Copyright 2025 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""Import low-level interface helpers into Python for Max287x chips."""

from ... import libpyuhd as lib

# Disable PyLint because we want to make it look like the following classes are
# defined in Python, but they're just renames of lib types. They therefore
# follow name conventions for Python classes, not for global constants.
# pylint: disable=invalid-name
Max2871Iface = lib.usrp.chips.max287x.max_2871
OutputPower = lib.usrp.chips.max287x.output_power
AuxOutputPower = lib.usrp.chips.max287x.aux_output_power
MuxoutMode = lib.usrp.chips.max287x.muxout_mode
ChargePumpCurrent = lib.usrp.chips.max287x.charge_pump_current
LdPinMode = lib.usrp.chips.max287x.ld_pin_mode
RFOutPort = lib.usrp.chips.max287x.rf_output_port
# pylint: enable=invalid-name



#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
Implements version check for a windows file including version comparison functionality.
"""

from ctypes import ( POINTER, byref , c_char, c_uint, cast, pointer, sizeof, Structure, windll, WinError)
from ctypes.wintypes import (
    BOOL, DWORD, LPCVOID, LPCWSTR, LPVOID
)

# Contains version information for a windows file to be accessed by windll.version.* Win32 API functions.
class VS_FIXEDFILEINFO(Structure):
    _fields_ = [
        ("dwSignature", DWORD),  # will be 0xFEEF04BD
        ("dwStrucVersion", DWORD),
        ("dwFileVersionMS", DWORD),
        ("dwFileVersionLS", DWORD),
        ("dwProductVersionMS", DWORD),
        ("dwProductVersionLS", DWORD),
        ("dwFileFlagsMask", DWORD),
        ("dwFileFlags", DWORD),
        ("dwFileOS", DWORD),
        ("dwFileType", DWORD),
        ("dwFileSubtype", DWORD),
        ("dwFileDateMS", DWORD),
        ("dwFileDateLS", DWORD)
    ]


def get_windll_version(filename):
    """
    Retrieve the version information of a specified file.
    Args:
        filename (str): The path to the file for which version information is to be retrieved.
    Returns:
        str: The version of the file in the format "major.minor.build.revision".
    Raises:
        WinError: If the operating system cannot retrieve version information for the file.
    """

    # Check whether the operating system can retrieve version information for the file. If not, raise an error.
    dwLen = windll.version.GetFileVersionInfoSizeW(filename, None)
    if not dwLen:
        raise WinError()
    
    # Retrieve version information for the file and store it into lpData buffer
    lpData = (c_char * dwLen)()
    if not windll.version.GetFileVersionInfoW(filename, 0, sizeof(lpData), lpData):
        raise WinError()
    
    # Retrieve version information from lpData
    uLen = c_uint()
    lpffi = POINTER(VS_FIXEDFILEINFO)()
    lplpBuffer = cast(pointer(lpffi), POINTER(LPVOID))
    if not windll.version.VerQueryValueW(lpData, u"\\", lplpBuffer, byref(uLen)):
        raise WinError()
    
    # dwFileVersion defines the binary version number for the file. The version consists of two 32-bit integers, 
    # defined by four 16-bit integers.
    ffi = lpffi.contents
    major = ffi.dwFileVersionMS >> 16
    minor = ffi.dwFileVersionMS & 0xFFFF
    build = ffi.dwFileVersionLS >> 16
    revision = ffi.dwFileVersionLS & 0xFFFF
    return f"{major}.{minor}.{build}.{revision}"


def compare_versions(version1, version2):
    """
    Compare two version strings.
    This function compares two version strings in the format 'X.Y.Z' where X, Y, and Z are integers.
    It considers only the first three components of the version strings.
    Parameters:
    version1 (str): The first version string to compare.
    version2 (str): The second version string to compare.
    Returns:
    int: -1 if version1 is less than version2, 1 if version1 is greater than version2, 0 if they are equal.
    """
    #print(f"compare {version1} against {version2}")

    # Extract the first three digits
    v1 = (version1.split('.'))[:3]
    v2 = (version2.split('.'))[:3]
    
    # Convert to tuple of integers for comparison
    v1_tuple = tuple(map(int, v1))
    v2_tuple = tuple(map(int, v2))
    
    # return the comparison result
    if v1_tuple < v2_tuple:
        return -1
    elif v1_tuple > v2_tuple:
        return 1
    else:
        return 0



"""RFNoC Modtool: Template wrapper.

This is a wrapper around the Mako templating engine.
"""

#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import mako.template


def render_wire_width(width, pad=8):
    """Render a wire's width (e.g., 8 -> [7:0]).

    If a pad value is provided, it will make the total width of the string as
    wide as the pad value.

    Examples:
    >>> render_wire_width(8, 10)
    "[     7:0]"
    >>> render_wire_width("WIDTH", 10) # Does not fit into 10 characters, will be longer
    "[WIDTH-1:0]"
    >>> len(render_wire_width(8, 10)) == 10
    True
    >>> render_wire_width(1, 4) # Single-bit wires don't need a width
    "    "

    Arguments:
        width: The wire width. If this is a string, it is interpreted as a
               Verilog constant and added verbatim.
        pad: Integer value, length of resulting string.
    """
    if isinstance(width, int):
        start_idx = width - 1
    else:
        start_idx = f"{width}-1"
    if start_idx == 0:
        return "".rjust(pad)
    range_str = f"{start_idx}:0".rjust(pad - 2)
    return f"[{range_str}]"


def quote(s):
    """Return a quoted string.

    The intention is to use this in templates to quote strings, such that they
    have quotes around them in the final output.
    """
    return f"'\"{s}\"'"


class Template(mako.template.Template):
    """Wrapper around the Mako template engine."""

    def __init__(self, *args, **kwargs):
        """Initialize the template.

        Differences to mako.template.Template:
        - strict_undefined is True by default
        """
        if "strict_undefined" not in kwargs:
            kwargs["strict_undefined"] = True
        super().__init__(*args, **kwargs)

    def render(self, **kwargs):
        """Render the template.

        This is identical to the parent class, but with the following extensions:

        - A 'quote' filter is provided. It is aliased to 'q'.
        """
        kwargs["quote"] = quote
        kwargs["q"] = quote
        return super().render(**kwargs)



"""Image builder submodule.

Copyright 2024 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later

This contains the big ImageBuilderConfig class, which is the data structure
that is passed to the templates.
"""



#!/usr/bin/env python3
#
# Copyright 2020 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""Utility to run power calibrations with USRPs.
"""

import argparse
import math
import pickle
import sys
import time

import uhd


def parse_args():
    """Parse args and return args object."""
    parser = argparse.ArgumentParser(
        description="Run power level calibration for supported USRPs",
    )
    parser.add_argument(
        "--args",
        default="",
        help="USRP Device Args",
    )
    parser.add_argument(
        "-d",
        "--dir",
        default="tx",
        help="Direction: Must be either rx or tx. From the perspective of the "
        "device: rx == receive power levels, tx == transmit power levels.",
        choices=["tx", "rx"],
    )
    parser.add_argument(
        "--start",
        help="Start Frequency. Defaults to lowest available frequency on device. "
        "Note that this is only a hint for the device object, which can choose "
        "to override this value.",
        type=float,
    )
    parser.add_argument(
        "--stop",
        help="Stop Frequency. Defaults to highest available frequency on device. "
        "Note that this is only a hint for the device object, which can choose "
        "to override this value.",
        type=float,
    )
    parser.add_argument(
        "--step",
        help="Frequency Step. Defaults are device-specific. "
        "Note that this is only a hint for the device object, which can choose "
        "to override this value. Devices can also measure at non-regular "
        "frequencies, e.g., to more accurately cover differences between bands.",
        type=float,
    )
    parser.add_argument(
        "--gain-step",
        type=float,
        default=1,
        help="Gain Step (dB). Defaults to 1 dB. "
        "Note that this is only a hint for the device object, which can choose "
        "to override this value. Devices can also measure at non-regular "
        "gain intervals.",
    )
    parser.add_argument(
        "--lo-offset",
        type=float,
        help="LO Offset. This gets applied to every tune request. Note that for "
        "TX measurements, there is also an offset applied by --tone-freq. "
        "The default value is device-dependent.",
    )
    parser.add_argument(
        "--amplitude",
        type=float,
        default=1.0 / math.sqrt(2),
        help="Amplitude of the tone that is generated for tx measurements. "
        "Default is 1/sqrt(2), or -3 dBFS.",
    )
    parser.add_argument(
        "--attenuation",
        type=float,
        default=0.0,
        help="Amount of attenuation between measurement device and DUT. This will "
        "be accounted for by simple addition, it is treated like a measurement error. "
        "The argument is generally positive, e.g. 30 means 30 dB of attenuation.",
    )
    parser.add_argument(
        "--tone-freq",
        type=float,
        default=1e6,
        help="Frequency of the tone that is generated for Tx measurements. This "
        "has the same effect as setting an LO offset, except in software.",
    )
    parser.add_argument(
        "--antenna",
        default="*",
        help="Select antenna port. A value of '*' means that the calibration "
        "will be repeated on all appropriate antenna ports.",
    )
    parser.add_argument(
        "--channels",
        default="*",
        help="Select channel. A value of '*' means that the calibration "
        "will be repeated on all appropriate channels.",
    )
    parser.add_argument(
        "--meas-dev", default="manual", help="Type of measurement device that is used"
    )
    parser.add_argument(
        "-o",
        "--meas-option",
        default=[],
        action="append",
        help="Options that are passed to the measurement device",
    )
    parser.add_argument(
        "--switch", default="manual", help="Type of switch to be used to connect antennas"
    )
    parser.add_argument(
        "--switch-option", default=[], action="append", help="Options that are passed to the switch"
    )
    parser.add_argument(
        "-r", "--rate", type=float, help="Sampling rate at which the calibration is performed"
    )
    parser.add_argument(
        "--store",
        metavar="filename.pickle",
        help="If provided, will store intermediate cal data. This can be analyzed "
        "separately, or loaded into the tool with --load.",
    )
    parser.add_argument(
        "--load",
        metavar="filename.pickle",
        help="If provided, will load intermediate cal data instead of running a " "measurement.",
    )
    return parser.parse_args()


def sanitize_args(usrp, args, default_rate):
    """Check the args against the USRP object."""
    assert (
        usrp.get_num_mboards() == 1
    ), "Power calibration tools are designed for a single motherboard!"
    available_chans = getattr(usrp, "get_{}_num_channels".format(args.dir))()
    if args.channels == "*":
        # * means all channels
        channels = list(range(available_chans))
    else:
        try:
            channels = [int(c) for c in args.channels.split(",")]
        except ValueError:
            raise ValueError("Invalid channel list: {}".format(args.channels))
        for chan in channels:
            assert chan in range(
                available_chans
            ), "ERROR: Invalid channel: {}. Should be in 0..{}.".format(chan, available_chans)
    print("=== Calibrating for channels:", ", ".join([str(x) for x in channels]))
    available_ants = getattr(usrp, "get_{}_antennas".format(args.dir))()
    if args.antenna == "*":
        invalid_antennas = ("CAL", "LOCAL", "CAL_LOOPBACK", "TERMINATION")
        antennas = [x for x in available_ants if x not in invalid_antennas]
    else:
        try:
            antennas = args.antenna.split(",")
        except ValueError:
            raise ValueError("Invalid antenna list: {}".format(args.antenna))
        for ant in antennas:
            assert ant in available_ants, "Invalid antenna: {}. Should be in {}.".format(
                ant, available_ants
            )
    print("=== Calibrating for antennas:", ", ".join([str(x) for x in antennas]))
    rate = args.rate or default_rate
    getattr(usrp, "set_{}_rate".format(args.dir))(rate)
    actual_rate = getattr(usrp, "get_{}_rate".format(args.dir))()
    print(
        "=== Requested sampling rate: {} Msps, actual rate: {} Msps".format(
            rate / 1e6, actual_rate / 1e6
        )
    )
    if args.dir == "tx" and abs(args.tone_freq) > actual_rate:
        raise ValueError(
            "The TX tone frequency offset of {} kHz is greater than the sampling rate.".format(
                args.tone_freq / 1e3
            )
        )
    return channels, antennas, actual_rate


def init_results(pickle_file):
    """Initialize results from pickle file, or empty dict."""
    if pickle_file is None:
        return {}
    with open(pickle_file, "rb") as results_file:
        return pickle.load(results_file)


class CalRunner:
    """Executor of the calibration routines."""

    def __init__(self, usrp, usrp_cal, meas_dev, args):
        """Initialize the cal runner."""
        self.usrp = usrp
        self.usrp_cal = usrp_cal
        self.meas_dev = meas_dev
        self.dir = args.dir
        self.tone_offset = args.tone_freq if args.dir == "tx" else 0.0
        self.lo_offset = args.lo_offset if args.lo_offset else usrp_cal.lo_offset
        if self.lo_offset:
            print("=== Using USRP LO offset: {:.2f} MHz".format(self.lo_offset / 1e6))

    def run(self, chan, freq):
        """Run all cal steps for a single frequency."""
        print("=== Running calibration at frequency {:.3f} MHz...".format(freq / 1e6))
        tune_req = uhd.types.TuneRequest(freq, self.lo_offset)
        getattr(self.usrp, "set_{}_freq".format(self.dir))(tune_req, chan)
        time.sleep(self.usrp_cal.tune_settling_time)
        actual_freq = getattr(self.usrp, "get_{}_freq".format(self.dir))(chan)
        if abs(actual_freq - freq) > 1.0:
            print(
                "WARNING: Frequency was coerced from {:.2f} MHz to {:.2f} MHz!".format(
                    freq / 1e6, actual_freq / 1e6
                )
            )
        self.meas_dev.set_frequency(actual_freq + self.tone_offset)
        getattr(self.usrp_cal, "run_{}_cal".format(self.dir))(freq)


def main():
    """Go, go, go!"""
    args = parse_args()
    print("=== Detecting USRP...")
    usrp = uhd.usrp.MultiUSRP(args.args)
    print("=== Measurement direction:", args.dir)
    print("=== Initializing measurement device...")
    meas_dev = uhd.usrp.cal.get_meas_device(args.dir, args.meas_dev, args.meas_option)
    meas_dev.power_offset = args.attenuation
    # If we're transmitting, then we need to factor in the "attenuation" from us
    # not transmitting at full scale
    if args.dir == "tx":
        meas_dev.power_offset -= 20 * math.log10(args.amplitude)
    print("=== Initializing port connector...")
    switch = uhd.usrp.cal.get_switch(args.dir, args.switch, args.switch_option)
    print("=== Initializing USRP calibration object...")
    usrp_cal = uhd.usrp.cal.get_usrp_calibrator(
        usrp,
        meas_dev,
        args.dir,
        gain_step=args.gain_step,
    )
    channels, antennas, rate = sanitize_args(usrp, args, usrp_cal.default_rate)
    results = init_results(args.load)
    usrp_cal.init(
        rate=rate,
        tone_freq=args.tone_freq,
        amplitude=args.amplitude,
    )
    print("=== Launching calibration...")
    cal_runner = CalRunner(usrp, usrp_cal, meas_dev, args)
    for chan in channels:
        if chan not in results:
            results[chan] = {}
        for ant in antennas:
            if ant in results[chan]:
                print("=== Using pickled data for channel {}, antenna {}.".format(chan, ant))
                continue
            print("=== Running calibration for channel {}, antenna {}.".format(chan, ant))
            # Set up all the objects
            getattr(usrp, "set_{}_antenna".format(args.dir))(ant, chan)
            switch.connect(chan, ant)
            usrp_cal.update_port(chan, ant)
            freqs = usrp_cal.init_frequencies(args.start, args.stop, args.step)
            usrp_cal.start()  # This will activate siggen
            # Now calibrate
            for freq in freqs:
                try:
                    cal_runner.run(chan, freq)
                except RuntimeError as ex:
                    print("ERROR: Stopping calibration due to exception: {}".format(str(ex)))
                    usrp_cal.stop()
                    return 1
            # Store results for pickling and shut down for next antenna port
            results[chan][ant] = usrp_cal.results
            usrp_cal.stop()  # This will deactivate siggen and store the data
    if args.store:
        print("=== Storing pickled calibration data to {}...".format(args.store))
        with open(args.store, "wb") as results_file:
            pickle.dump(results, results_file)
    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except (RuntimeError, ValueError) as ex:
        print("ERROR:", str(ex))
        sys.exit(1)



#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
""" @package usrp_clock
Python UHD module containing the MultiUSRPClock
"""

from .multi_usrp_clock import MultiUSRPClock



"""RFNoC Utilities: Logging.

SPDX-License-Identifier: GPL-3.0-or-later
"""

import logging
import sys


class ColorFormatter(logging.Formatter):
    """Logging Formatter to add colors and icons."""

    RESET = 0
    BOLD = 1
    DIM = 2
    GREY = "38;20"
    BLACK = "30;20"
    YELLOW = "33;20"
    RED = "31;20"
    BRIGHTRED = 91
    BRIGHT = 99

    def c(colors):  # noqa -- should be staticmethod but that requires Python 3.10
        """Format escape sequence from list of colors."""
        return f"\x1b[{';'.join(str(c) for c in colors)}m"

    debug_color = c([DIM])
    info_color = c([BRIGHT, BOLD])
    warning_color = c([YELLOW])
    error_color = c([BRIGHTRED])
    crit_color = c([RED, BOLD])
    reset = c([RESET])

    FORMATS = {
        logging.DEBUG: debug_color + "[debug] %(message)s" + reset,
        logging.INFO: info_color + "%(message)s" + reset,
        logging.WARNING: warning_color + "⚠   %(message)s" + reset,
        logging.ERROR: error_color + "⛔   %(message)s" + reset,
        logging.CRITICAL: crit_color + "⛔   %(message)s" + reset,
    }

    def format(self, record):
        """Format a record the way we like it."""
        log_fmt = self.FORMATS.get(record.levelno)
        return logging.Formatter(log_fmt).format(record)


class SimpleFormatter(logging.Formatter):
    """Logging Formatter for non-interactive shells."""

    FORMATS = {
        logging.DEBUG: "[debug] %(message)s",
        logging.INFO: "%(message)s",
        logging.WARNING: "[warning] %(message)s",
        logging.ERROR: "[error] %(message)s",
        logging.CRITICAL: "[critical] %(message)s",
    }

    def format(self, record):
        """Format a record the way we like it."""
        log_fmt = self.FORMATS.get(record.levelno)
        return logging.Formatter(log_fmt).format(record)


def init_logging(color_mode="auto", log_level="info"):
    """Initialize the logging interface for RFNoC utilities."""
    use_color = (color_mode == "always") or (
        color_mode == "auto" and sys.__stdout__.isatty() and sys.__stderr__.isatty()
    )
    handler = logging.StreamHandler()
    handler.setFormatter(ColorFormatter() if use_color else SimpleFormatter())
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    if log_level:
        logging.root.setLevel(log_level.upper())



"""
Copyright (c) 2023 Ettus Research, A National Instruments Brand

SPDX-License-Identifier: GPL-3.0-or-later
"""

from .command import BaseCommand

MPM_RPC_PORT = 49601

class ResetCommand(BaseCommand):
    """
    Command that resets the specified subcomponents of the USRP device
    """
    @classmethod
    def add_parser(cls, parser):
        """
        Allow -mpm as an argument, other subcomponents and full device reset
        not supported right now
        """
        subparser = parser.add_parser(cls.command_name(),
                help="reset the specified subcomponent(s) on the USRP device")
        subparser.add_argument("-mpm", const="mpm", action="store_const",
                help="reset MPM on the USRP device")

    def is_multi_device_capable(self):
        """
        Can handle multiple USRPs
        """
        return True

    def run(self, usrps, args):
        """
        Reset the specified subcomponents for the USRP device(s)
        """
        args_list = self._to_arg_list(args)
        if not args_list:
            print("No subcomponent specified for reset, reset did nothing")
            return
        for usrp in usrps:
            if 'mpm' in args_list:
                import msgpackrpc
                try:
                    if not 'mgmt_addr' in usrp:
                        # For devices that don't support MPM (e.g. x310)
                        print("Device does not have mgmt_addr, MPM not reset")
                        continue
                    # Increase timeout from default, since MPM restarting can take longer 
                    # than 10 seconds on some devices
                    client = msgpackrpc.Client(msgpackrpc.Address(usrp['mgmt_addr'], MPM_RPC_PORT),
                                               timeout=20)
                    token = client.call('claim', 'usrpctl')
                    print(f"Resetting MPM for device at {usrp['mgmt_addr']}...")
                    client.call('reset_timer_and_mgr', token)
                    print("MPM reset successfully")
                except Exception as e:
                    print(f"Resetting MPM was unsuccessful: {e}")



"""Image builder connections module.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later

Functions, helpers, and other code to handle connections in the image builder.
"""

import copy
import re
import sys

from .common import NONE_PORT


def con2str(con):
    """Pretty-print a single connection."""
    return f"{con['srcblk']}:{con['srcport']} → {con['dstblk']}:{con['dstport']}"


def _check_duplicate_connections(config):
    """Check that ports don't have duplicate connections.

    Checks that a port that is referenced in a connection is not referenced in
    any other connection. The exception are broadcasters, which can have multiple
    destinations.

    As a reminder, the connections object looks something like this:
    [
        {'srcblk': src1, 'srcport': srcport1, 'dstblk': dst1, 'dstport': dstport1},
        {'srcblk': src2, 'srcport': srcport2, 'dstblk': dst2, 'dstport': dstport2},
        ...
    ]

    A duplicate connection is any tuple of connections where either the srcblk
    and srcport or the dstblk and dstport are the same (again, with the exception
    of broadcasters).
    """
    for con_idx, con in enumerate(config.connections):
        for s_d in ("src", "dst"):

            def is_broadcast(c):
                return s_d == "src" and c["srctype"] == "broadcaster"

            def port_match(c):
                return con[f"{s_d}blk"] == c[f"{s_d}blk"] and con[f"{s_d}port"] == c[f"{s_d}port"]

            dupes = [
                other
                for other in config.connections[con_idx + 1 :]
                if port_match(other) and not is_broadcast(other)
            ]
            if dupes:
                error = (
                    f"The {'source' if s_d == 'src' else 'destination'} port {con[f'{s_d}blk']}:{con[f'{s_d}port']} is "
                    f"connected to multiple {'destinations' if s_d == 'src' else 'sources'} via the following connections:\n"
                    f"{con2str(con)}\n"
                )
                for dupe in dupes:
                    error += con2str(dupe) + "\n"
                config.log.error(error)


def _check_io_port_compatibility(con, log):
    """Check if source and destination IO ports are compatible."""
    if con["src_iosig"].get("type") != con["dst_iosig"].get("type"):
        log.error(
            f"IO port type mismatch: {con['srcblk']}.{con['srcport']} "
            f"(type: {con['src_iosig'].get('type')}) → "
            f"{con['dstblk']}.{con['dstport']} "
            f"(type: {con['dst_iosig'].get('type')})"
        )
    src_ports = con["src_iosig"].get("wires", [])
    dst_ports = con["dst_iosig"].get("wires", [])
    if len(src_ports) != len(dst_ports):
        log.warning(
            f"Connection {con2str(con)} has a port number mismatch. "
            f"Counting {len(src_ports)} ports on source and {len(dst_ports)} ports on destination."
        )
    for src_port, dst_port in zip(src_ports, dst_ports):
        src_width = src_port.get("width", 1)
        dst_width = dst_port.get("width", 1)
        if src_width != dst_width:
            log.warning(
                f"Connection {con2str(con)} has a wire width mismatch. "
                f"Source port {src_port['name']} has width {src_width}, destination port {dst_port['name']} has width {dst_width}."
            )


def check_and_sanitize(config):
    """Check and sanitize connections of an image builder configuration.

    This will run through all the connections on a configuration object and
    check if they're valid, but it will also modify them:

    - Add annotations to the connections about domain, type, and IO signature
    - Provide error messages if not valid
    """
    failure = ""

    def sanitize_port(con, s_d, failure):
        """Unpack port names (separate slice from port name)."""
        port_match = re.match(r"^([a-z0-9_]+)(?:\[([^]])\])?$", con[f"{s_d}port"])
        if not port_match:
            failure += f"Invalid port name: {con[f'{s_d}port']}\n"
            return con
        con[f"{s_d}port"], con[f"{s_d}slice"] = port_match.groups()
        return con

    def check_port_is_valid(blk_key, blk, port, s_d):
        """Return true if the port is valid for the block."""
        io_ports = getattr(blk, "io_ports", {})
        valid_io_port_drives = ("master", "broadcaster") if s_d == "src" else ("slave", "listener")
        return (
            ((blk_key, port, "output" if s_d == "src" else "input") in config.block_ports)
            or (io_ports.get(port, {}).get("drive") in valid_io_port_drives)
            or (port == NONE_PORT)
        )

    def get_ports(blk_key, blk, s_d):
        """Return list of all valid ports for a block."""
        data_ports = [
            p[1]
            for p in config.block_ports
            if p[0] == blk_key and p[2] == ("output" if s_d == "src" else "input")
        ]
        valid_io_port_drives = ("master", "broadcaster") if s_d == "src" else ("slave", "listener")
        io_ports = [
            k for k, v in getattr(blk, "io_ports", {}).items() if v["drive"] in valid_io_port_drives
        ]
        return ", ".join(data_ports + io_ports)

    ## Phase 1: We go through the list of connections and provide annotations
    for conn_idx, con in enumerate(config.connections):
        for s_d in ("src", "dst"):
            con = sanitize_port(con, s_d, failure)
        src_blk = config.get_module(con["srcblk"])
        if src_blk is None:
            failure += f"Source block '{con['srcblk']}' not found\n"
            continue
        dst_blk = config.get_module(con["dstblk"])
        if dst_blk is None:
            failure += f"Destination block '{con['dstblk']}' not found\n"
            continue
        if not check_port_is_valid(con["srcblk"], src_blk, con["srcport"], "src"):
            failure += f"Source port '{con['srcblk']}.{con['srcport']}' not found. Available ports on '{con['srcblk']}': {get_ports(con['srcblk'], src_blk, 'src')}\n"
            continue
        if not check_port_is_valid(con["dstblk"], dst_blk, con["dstport"], "dst"):
            failure += f"Destination port '{con['dstblk']}.{con['dstport']}' not found. Available ports on '{con['dstblk']}': {get_ports(con['dstblk'], dst_blk, 'dst')}\n"
            continue
        try:
            # If dummy connection, we just skip this.
            if con["srcport"] == NONE_PORT or con["dstport"] == NONE_PORT:
                pass
            # If this is a connection between blocks, or between stream endpoints and
            # blocks, then we can determine the connection type.
            elif (con["srcblk"], con["srcport"], "output") in config.block_ports and (
                con["dstblk"],
                con["dstport"],
                "input",
            ) in config.block_ports:
                con["srctype"] = "output"
                con["dsttype"] = "input"
            # If it's none of the above, it's either an IO port connection, or it's
            # not valid.
            elif (
                src_blk.io_ports.get(con["srcport"], {}).get("drive") == "master"
                and dst_blk.io_ports.get(con["dstport"], {}).get("drive") == "slave"
            ) or (
                src_blk.io_ports.get(con["srcport"], {}).get("drive") == "broadcaster"
                and dst_blk.io_ports.get(con["dstport"], {}).get("drive") == "listener"
            ):
                # In this case, we have an IO port. Make sure we have the IO
                # signature copied into the connection object.
                con["srctype"] = src_blk.io_ports[con["srcport"]]["drive"]
                con["dsttype"] = dst_blk.io_ports[con["dstport"]]["drive"]
                con["src_iosig"] = copy.deepcopy(src_blk.io_ports[con["srcport"]])
                con["dst_iosig"] = copy.deepcopy(dst_blk.io_ports[con["dstport"]])
                _check_io_port_compatibility(con, config.log)
            # If we have IO ports, but the previous clause didn't take, then
            # the connection endpoints are not compatible.
            elif (con["srcport"] in src_blk.io_ports) and (con["dstport"] in dst_blk.io_ports):
                con["srctype"] = src_blk.io_ports[con["srcport"]]["drive"]
                con["dsttype"] = dst_blk.io_ports[con["dstport"]]["drive"]
                failure += (
                    f"IO port {con['srcblk']}.{con['srcport']} "
                    f"({src_blk.io_ports[con['srcport']]['drive']}) "
                    f"cannot connect to {con['dstblk']}.{con['dstport']} "
                    f"({dst_blk.io_ports[con['dstport']]['drive']})\n"
                    f"Make sure io ports are connected master      (src) to slave    (dst)\n"
                    f"                              or broadcaster (src) to listener (dst)\n"
                )
            else:
                failure += "Unresolved connection: " + con2str(con) + "\n"
            config.connections[conn_idx] = con
        except AttributeError as ex:
            config.log.error(
                "Error parsing connection: %s. No io_ports attribute on source or dest:\n%s\n%s (%s)",
                con,
                src_blk,
                dst_blk,
                str(ex),
            )
            raise

    # Go through all the modules and check IO ports are connected
    for module_name, module in config.get_module_list("all").items():
        for io_port_name, io_port in module.io_ports.items():
            required = io_port.get("required")
            if required:
                is_connected = any(
                    con["srcblk"] == module_name
                    and con["srcport"] == io_port_name
                    or con["dstblk"] == module_name
                    and con["dstport"] == io_port_name
                    for con in config.connections
                )
                if not is_connected:
                    available_ports = config._get_available_ports(io_port)
                    msg = f"IO port {module_name}.{io_port_name} is not connected. "
                    if len(available_ports) == 1:
                        msg += (
                            f"Suggested connection: "
                            f"{available_ports[0][0]}.{available_ports[0][1]}\n"
                        )
                        msg += (
                            f"Add the following connection to the image core file to "
                            f"use this connection:\n"
                        )
                        if available_ports[0][2] == "src":
                            msg += (
                                f"{{ srcblk: {available_ports[0][0]}, "
                                f"srcport: {available_ports[0][1]}, dstblk: "
                                f"{module_name}, dstport: {io_port_name} }}"
                            )
                        else:
                            msg += (
                                f"{{ srcblk: {module_name}, srcport: {io_port_name}, "
                                f"dstblk: {available_ports[0][0]}, dstport: "
                                f"{available_ports[0][1]} }}"
                            )
                    elif len(available_ports) > 1:
                        msg += f"Available connections:\n"
                        for port in available_ports:
                            msg += f"    {port[0]}.{port[1]}\n"
                        msg += (
                            "Add a connection line to the image core file to enable "
                            "this connection:\n"
                        )
                        if available_ports[0][2] == "src":
                            msg += (
                                f"{{ srcblk: ..., srcport: ..., dstblk: "
                                f"{module_name}, dstport: {io_port_name} }}"
                            )
                        else:
                            msg += (
                                f"{{ srcblk: {module_name}, srcport: {io_port_name}, "
                                f"dstblk: ..., dstport: ... }}"
                            )
                    else:
                        msg += f"No available IO ports found!"
                    if required == "recommended":
                        config.log.warning(msg)
                    else:
                        failure += msg
    # Go through blocks and check all ports are connected
    for block_name, block in config.noc_blocks.items():
        for direction in ("inputs", "outputs"):
            for port_name in block["data"][direction]:
                if not any(
                    con["srcblk"] == block_name
                    and con["srcport"] == port_name
                    or con["dstblk"] == block_name
                    and con["dstport"] == port_name
                    for con in config.connections
                ):
                    config.log.warning("Block port %s.%s is not connected", block_name, port_name)
    # Drop empty connections
    config.connections = [
        c for c in config.connections if c["srcport"] != NONE_PORT and c["dstport"] != NONE_PORT
    ]
    # Now make sure there are no duplicate connections
    _check_duplicate_connections(config)

    if failure:
        config.log.error(
            "There are errors in the image core file's connection settings:\n" + failure
        )
        sys.exit(1)


def get_num_ports_from_data_port(port_info, block):
    """Resolve number of ports from data port info.

    Rules:
    - If num_ports is not in port_info, return 1
    - If it's an integer, return that.
    - If num_ports isn't an integer, it could be an expression
      using values from the parameters section (e.g.,
      NUM_PORTS*NUM_BRANCHES for a stream-splitting block).
      If the parameter doesn't resolve to an integer, treat it
      as an expression that needs to be evaluated, hopefully to
      an integer.
    """
    if "num_ports" not in port_info:
        return 1
    num_ports = port_info["num_ports"]
    if isinstance(num_ports, int):
        return num_ports

    # Create a regex to find identifiers.
    regex_ident = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")

    # Get a list of all identifiers in the num_ports
    # expression and iterate over them all
    idents = re.finditer(regex_ident, num_ports)
    for ident in idents:
        # If the identifier represents a valid parameter
        # in the block, replace the identifier text with
        # the value of the parameter. If no matching
        # parameter is found, just leave the text in
        # place. That may result in an exception being
        # thrown from eval(), but we'll catch it and
        # report an error a bit later on.
        if ident[0] in block["parameters"]:
            val = str(block["parameters"][ident[0]])
            num_ports = re.sub(ident[0], val, num_ports)

    # Now, with identifiers resolved to parameter values,
    # attempt to evaluate the expression. If eval() fails,
    # we'll catch the exception, num_ports will remain non-
    # integral, and the if statement after the exception
    # is caught will inform the user.
    try:
        return eval(num_ports, {}, {})
    except:  # noqa: E722
        return None



"""Image builder: Helpers for interacting with YAML sources.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later

This module contains methods, helpers, utilities to deal with handling YAML
files for RFNoC. This includes knowledge about paths, specific contents, and
RFNoC-specific extensions.
"""

import io
import json
import logging
import os
import re
import sys
from collections import OrderedDict
from collections.abc import Mapping

from ruamel import yaml

from .utils import merge_dicts

# Allow jsonschema import to fail. If not available no schema validation will
# be done (but warning will be printed for each skipped validation).
try:
    import jsonschema
except ImportError:
    logging.warning(
        "Module jsonschema is not installed. Configuration files "
        "will not be validated against their schema."
    )


# List of blocks that have been replaced (old name : new name)
deprecated_block_yml_map = {
    "radio_1x64.yml": "radio.yml",
    "radio_2x64.yml": "radio.yml",
    "axi_ram_fifo_2x64.yml": "axi_ram_fifo.yml",
    "axi_ram_fifo_4x64.yml": "axi_ram_fifo.yml",
}

# List of IO signature types that have been replaced (old name : new name)
deprecated_port_type_map = {
    "ctrl_port": "ctrlport",
    "time_keeper": "timekeeper",
    "radio_1x32": "radio",
    "radio_2x32": "radio",
    "radio_8x32": "radio",
    "x300_radio": "radio",
}

# List of port names that have been replaced (old name : new name)
deprecated_port_name_map = {
    "x300_radio": "radio",
    "radio_iface": "radio",
    "x300_radio0": "radio0",
    "x300_radio1": "radio1",
    "radio_ch0": "radio0",
    "radio_ch1": "radio1",
    "ctrl_port": "ctrlport",
    "ctrlport_radio": "ctrlport",
    "timekeeper": "time",
    "time_keeper": "time",
}


# pylint: disable=too-few-public-methods
class IOConfig(Mapping):
    """Class containing configuration from a yml file.

    Each top level entry is translated into a class member variable. If the
    configuration contains an io_ports section the ports get a wire list which
    is derived from the signature file. This allows easier processing of IO
    ports in the mako templates and failures from yml configuration files fail
    in this script rather than during template processing which is easier to
    track and debug.
    """

    def __init__(self, config, signatures):
        """Read configuration from config dictionary."""
        # TODO: Is this guaranteed ordered?
        self.__dict__.update(**config)
        if hasattr(self, "io_ports"):
            expand_io_port_desc(getattr(self, "io_ports"), signatures)

    def __iter__(self):
        """Return iterator (so we can use this as a mapping)."""
        return iter(self.__dict__)

    def __getitem__(self, key):
        """Implement getitem (so we can use this as a mapping)."""
        return self.__dict__[key]

    def __len__(self) -> int:
        """Implement len (so we can use this as a mapping)."""
        return len(self.__dict__)


# pylint: enable=too-few-public-methods

USRP3_TOP_DIR = os.path.join("usrp3", "top")

# Subdirectory for the core YAML files
RFNOC_CORE_DIR = os.path.join("rfnoc", "core")


def get_top_path(fpga_root):
    """Return the path where FPGA top level sources reside."""
    return os.path.join(fpga_root, USRP3_TOP_DIR)


def get_core_config_path(config_path):
    """Return the path where core configuration files are stored."""
    return os.path.join(config_path, RFNOC_CORE_DIR)


def find_file(file_name, base_path, recurse=True):
    """Search for file recursively.

    Recursive search for a file. Only looks for a file with appropriate
    name without checking for content or accessibility.

    :param file_name: name of file to search for
    :param base_path: root path to start search in
    :return: full path to schema file if found, None else
    """
    if not recurse:
        if os.path.isfile(os.path.join(base_path, file_name)):
            return os.path.join(base_path, file_name)
    for root, _, _ in os.walk(base_path):
        filename = os.path.join(root, file_name)
        if os.path.isfile(filename):
            return filename
    return None


def validate_config(config, config_path):
    """Try to validate config.

    config contains a configuration loaded from a yaml file. config is assumed
    to be a dictionary which contains a key 'schema' which determines
    which schema to validate against. The schema (json formatted) needs to be
    located in config_path or any sub folder of it.
    If "jsonschema" module cannot be loaded validation is skipped and config
    is assumed to be valid. This way a configuration can also be loaded if
    "jsonschema" is not available.
    The method raises ValueError if no schema is defined in config or the
    schema defined in config cannot be found. The validation itself may throw
    a jsonschema.exceptions.ValidationError if config does not confirm to its
    schema.
    :param config: a dictionary to validate (loaded from yaml file).
    :param config_path: a path holding schema definitions
    """
    if "jsonschema" not in sys.modules:
        logging.warning("Skip schema validation (missing module jsonschema).")
        return

    if "schema" not in config:
        raise ValueError("Missing schema in configuration.")

    schema_name = config["schema"]
    logging.debug("Validating against schema %s...", schema_name)

    schema_file = find_file(f"{schema_name}.json", config_path)
    if not schema_file:
        raise ValueError(f"Unknown schema: '{schema_name}'.")

    logging.debug("Using schema file %s.", schema_file)

    with open(schema_file, encoding="utf-8") as stream:
        jsonschema.validate(instance=config, schema=json.load(stream))
        logging.debug("Configuration successful validated.")


def load_config_validate(config_file, config_path, allow_inherit=True):
    """Load configuration files.

    Wrapper method to unify loading of configuration files.
    Beside loading the configuration (yaml file format) itself from config_file
    this method also validates the configuration against a schema. The root
    element of the configuration must load to a dictionary which contains a
    "schema" key. The value of schema points to a file named {value}.json which
    must be located in config_path or one of its sub folders.
    .. seealso:: validate_config.
    :param config_file: configuration to load
    :param config_path: root path of schema definition files
    :return:
    """
    logging.debug("Loading configuration %s...", config_file)
    with open(config_file, encoding="utf-8") as stream:
        rt_yaml = yaml.YAML(typ="rt")
        config = rt_yaml.load(stream)
        if allow_inherit and "inherit" in config:
            inherits = config["inherit"]
            if not isinstance(inherits, list):
                inherits = [inherits]
            for inherit in inherits:
                logging.debug("Image core file inherits from %s...", inherit)
                parent_file = (
                    find_file(inherit, "", False)
                    or find_file(inherit, os.path.dirname(config_file), False)
                    or find_file(inherit, get_core_config_path(config_path))
                )
                logging.debug("Found parent file: %s", parent_file)
                if not parent_file:
                    logging.error(
                        "Cannot find parent file %s requested by %s!", inherit, config_file
                    )
                    sys.exit(1)
                parent_config = load_config_validate(
                    parent_file,
                    config_path,
                    allow_inherit=True,  # Parent files can also inherit, why not
                )
                config = merge_dicts(parent_config, config)
            config.pop("inherit")
        logging.debug("Configuration successful loaded.")
        validate_config(config, config_path)
        return config


def load_config(filename):
    """Load YAML configuration from filename.

    Configuration files are searched in folder returned by get_get_config_path.
    This method logs error and exits on IO failure

    :param filename: yml configuration to load
    :return: IO signatures as dictionary
    """
    dirname, basename = os.path.split(filename)
    try:
        with open(filename, encoding="utf-8") as stream:
            logging.debug("Using %s from %s.", basename, os.path.normpath(dirname))
            rt_yaml = yaml.YAML(typ="rt")
            config = rt_yaml.load(stream)
        return config
    except IOError:
        logging.error("%s misses %s", os.path.normpath(dirname), basename)
        sys.exit(1)


def device_config(config_path, device):
    """Load device config from bsp.yml.

    Location of bsp.yml is derived from the device chosen in the arguments

    :param config_path: location of core configuration files
    :param device: device to build for
    :return: device configuration as dictionary
    """
    return load_config(os.path.join(config_path, f"{device.lower()}_bsp.yml"))


def io_signatures(config_path, *modules):
    """Load IO signatures from io_signatures.yml and modules.

    First IO signatures from config_path are loaded. All signatures from
    the config_path are marked as `core` in the `origin` key. Next all
    modules are scanned for a `io_signatures` key on the top level definition
    of the module. If one is found the final signature definitions is extended
    by these signatures and their origin is marked by the name of the module.

    :param config_path: location of core configuration files
    :param modules: a list of modules that might contain additional IO signatures
    :return: IO signatures as dictionary
    """
    result = load_config(os.path.join(config_path, "io_signatures.yml"))
    for key in result:
        result[key]["origin"] = "core"

    for module in modules:
        for module_file, module_data in module.items():
            for signature, data in module_data.get("io_signatures", {}).items():
                if signature in result:
                    # temporary take the same origin for compare in next line to succeed
                    data["origin"] = result[signature]["origin"]
                    if result[signature] == data:
                        logging.debug(
                            "Identical signature '%s' from '%s' " "already defined in '%s'. Skip",
                            signature,
                            module_file,
                            data["origin"],
                        )
                        continue

                    error = (
                        f"Redefinition of signature: '{signature}' in "
                        f"'{module_file}'. Different definition already "
                        f"in '{data['origin']}'"
                    )
                    logging.error(error)
                    logging.error(
                        "Existing definition from '%s': %s", data["origin"], result[signature]
                    )
                    logging.error("New definition from '%s': %s", module_file, data)
                    raise ValueError(error)

                data["origin"] = module_file
                result[signature] = data

    logging.debug("Loaded %d IO signatures", len(result))
    for signature in result:
        logging.debug("\t%s [%s]", signature, result[signature]["origin"])

    return result


def read_yaml_definitions(*paths):
    """Non-recursively search all paths for YAML definitions.

    :param paths: paths to be searched
    :return: dictionary of noc blocks. Key is filename of the block, value
             is an IOConfig object
    """
    blocks = OrderedDict()
    for path in paths:
        for (
            root,
            _,
            files,
        ) in os.walk(path):
            for filename in files:
                if re.match(r".*\.ya?ml$", filename):
                    with open(os.path.join(root, filename), encoding="utf-8") as stream:
                        yaml_rt = yaml.YAML(typ="rt")
                        data = yaml_rt.load(stream)
                        if filename in deprecated_block_yml_map:
                            logging.warning(
                                "Skipping deprecated block description " "%s (%s).",
                                filename,
                                os.path.normpath(root),
                            )
                        else:
                            logging.debug("Adding file %s (%s).", filename, os.path.normpath(root))
                            blocks[filename] = data
    return blocks


def resolve_io_signatures(blocks, signatures, require_schema):
    """Resolve the IO signatures in blocks.

    Resolving the signatures is postponed from `read_yaml_definitions`
    because each yaml file read could potentially contain additional
    IO signatures that needs to be loaded first.

    :param blocks: OrderedDict of definitions
    :param signatures: signature passed to IOConfig initialization
    :param reqire_schema: only blocks with this schema will be resolved
    """
    for name, data in blocks.items():
        if data.get("schema") == require_schema:
            blocks[name] = IOConfig(data, signatures)


def expand_io_port_desc(io_ports, signatures):
    """Expand wires on IO port info.

    Add a wires entry to each io port dictionary entry which contains a
    complete list of wires for the specific port according to the information
    in signature file. Each wire entry contains:
    * fully resolved wire name (wire name as in signature or replaced name
      with respect to regular expression if some is given, regular expression
      should use back references to retain original wire name).
    * width in bits
    * direction as input/output depending on whether the port is a
      master/broadcaster or slave/listener and the wire is described as from
      or to master
    :param io_ports: io port dictionary from yml configuration
    :param signatures: signature description from yml configuration
    :return: None
    """
    for io_port in io_ports.values():
        wires = []
        if io_port["type"] in deprecated_port_type_map:
            logging.warning(
                "The IO signature type '%s' has been deprecated. "
                "Please update your block YAML to use '%s'.",
                io_port["type"],
                deprecated_port_type_map[io_port["type"]],
            )
            io_port["type"] = deprecated_port_type_map[io_port["type"]]
        if io_port["type"] not in signatures:
            logging.error("Unknown IO signature type '%s'.", io_port["type"])
            sys.exit(1)
        for signature in signatures[io_port["type"]]["ports"]:
            width = signature.get("width", 1)
            wire_type = signature.get("type", None)
            drive = io_port["drive"]
            direction = {
                "master": {"from-master": "input ", "to-master": "output"},
                "slave": {"from-master": "output", "to-master": "input "},
                "broadcaster": {None: "input "},
                "listener": {None: "output"},
            }[drive][wire_type]

            signature_name = signature["name"]
            if "rename" in io_port:
                signature_name = re.sub(
                    io_port["rename"]["pattern"], io_port["rename"]["repl"], signature_name, 1
                )

            wires.append({"direction": direction, "width": width, "name": signature_name})
            if "safe" in signature:
                wires[-1]["safe"] = signature["safe"]
        io_port["wires"] = wires
        io_port["parameters"] = {
            **signatures[io_port["type"]].get("parameters", {}),
            **io_port.get("parameters", {}),
        }


def write_yaml(config, destination):
    """Write a config object as a YAML file."""
    with open(destination, "w", encoding="utf-8") as out_file:
        # TODO remove superfluous comments
        rt_yaml = yaml.YAML(typ="rt")
        rt_yaml.default_flow_style = False
        rt_yaml.indent(mapping=4)
        rt_yaml.dump(config, out_file)


def read_yaml(yml_file):
    """Read a YAML file."""
    with open(yml_file, encoding="utf-8") as stream:
        rt_yaml = yaml.YAML(typ="rt")
        return rt_yaml.load(stream)


def reload_dict(cfg_object):
    """Save a dictionary to a YAML file and reload it."""
    # dump to string
    rt_yaml = yaml.YAML(typ="rt")
    rt_yaml.default_flow_style = False
    rt_yaml.indent(mapping=4)
    strdump = io.StringIO()
    rt_yaml.dump(cfg_object, strdump)
    # load from string
    strdump.seek(0)
    return rt_yaml.load(strdump)



#!/usr/bin/env python3
"""RFNoC Modtool: The tool to create and manipulate RFNoC OOT modules."""
#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import argparse
import glob
import os
import sys

from ruamel.yaml import YAML
from uhd import get_pkg_data_path

from .log import init_logging
from .step_executor import StepExecutor
from .utils import resolve


def get_command_repo_dir():
    """Return the path to the command repository directory (where the command YAMLS are stored)."""
    this_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(this_dir, "modtool_commands")


def collect_commands():
    """Load all available commands.

    The return value is a dictionary where the keys are the command names and
    the values are raw contents of the YAML as Python structures.
    """
    yamls = glob.glob(os.path.join(get_command_repo_dir(), "*.yml"))
    yaml = YAML()
    commands = {}
    for y in yamls:
        cmd_name = os.path.basename(y).replace(".yml", "")
        with open(y, "r", encoding="utf-8") as f:
            commands[cmd_name] = yaml.load(f)
    return commands


def parse_args(cmds):
    """Parse args.

    This will dynamically populate main- and subparsers from the YAML configs
    given in cmds.
    """
    arg_parser = argparse.ArgumentParser(description=__doc__)
    arg_parser.add_argument(
        "-C", "--directory", help="Change to this directory before running the command"
    )
    arg_parser.add_argument(
        "-l", "--log-level", default="INFO", help="Set the log level (default: INFO)"
    )
    subparsers = arg_parser.add_subparsers(dest="command", required=True)
    parser_list = []
    for cmd in cmds:
        new_parser = subparsers.add_parser(cmd, help=cmds[cmd]["help"])
        for arg_name, arg_info in cmds[cmd]["args"].items():
            name_or_flags = (
                f"{arg_name}" if "name_or_flags" not in arg_info else arg_info.pop("name_or_flags")
            )
            if not isinstance(name_or_flags, list):
                name_or_flags = [name_or_flags]
            # We don't want to use eval(), but for most types we need to do
            # just that, so manually map
            if "type" in arg_info:
                try:
                    arg_info["type"] = {
                        "str": str,
                        "int": int,
                        "float": float,
                    }[arg_info["type"]]
                except KeyError:
                    pass
            new_parser.add_argument(*name_or_flags, **arg_info)
        parser_list.append(new_parser)
    return arg_parser.parse_args()


def get_global_vars(pkg_data_dir):
    """Create a dictionary with global variables for the key resolution."""
    return {
        "RFNOC_PKG_DIR": pkg_data_dir,
        "CWD": os.getcwd(),
    }


def resolve_vars(cmd, global_vars, args):
    """Resolve all variables in the command."""
    for var_name, var_val in cmd.get("variables", {}).items():
        cmd["variables"][var_name] = resolve(
            var_val, args=args, **global_vars, **cmd.get("variables", {})
        )
    return cmd


def check_valid_oot_dir(oot_dir):
    """Check if the given directory contains a valid OOT module."""
    return os.path.isfile(os.path.join(oot_dir, "CMakeLists.txt")) and os.path.isdir(
        os.path.join(oot_dir, "rfnoc")
    )


def main():
    """Run main rfnoc_modtool function."""
    pkg_data_dir = os.path.normpath(get_pkg_data_path())
    cmds = collect_commands()
    args = parse_args(cmds)
    init_logging(log_level=args.log_level)
    if args.directory:
        os.chdir(args.directory)
    cmd = cmds[args.command]
    global_vars = get_global_vars(pkg_data_dir)
    if not cmd.get("skip_identify_module", False):
        if not check_valid_oot_dir(os.getcwd()):
            print("Error: Not a valid OOT module directory")
            return 1
        oot_dir_name = os.path.split(os.getcwd())[-1]
        module_name = oot_dir_name.replace("rfnoc-", "")
        global_vars["MODULE_NAME"] = module_name
        global_vars["MODULE_NAME_FULL"] = oot_dir_name
        global_vars["MODULE_DIR"] = os.getcwd()
    cmd = resolve_vars(cmd, global_vars, args)
    executor = StepExecutor(global_vars, args, cmd)
    executor.run(cmd["steps"])
    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""RFNoC Modtool: Step Executor.

Contains the class that can run individual steps of a command script.
"""
#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import datetime
import fnmatch
import glob
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback
from argparse import Namespace
from pathlib import Path

import mako.lookup
from ruamel.yaml import YAML

from . import yaml_utils
from .template import Template, render_wire_width
from .utils import resolve


def get_file_list(**kwargs):
    """Extract a list of filenames from kwargs.

    The following rules apply:
    - If "glob" is present, use it to generate a list of filenames.
    - If "files" is present, use it as a list of filenames.
    - If "file" is present, use it as a single filename.
    """
    file_list = []
    if "glob" in kwargs and kwargs["glob"]:
        glob_pattern = kwargs["glob"]
        # FIXME: Maybe the user wants to glob somewhere else
        glob_pattern = os.path.join(os.getcwd(), glob_pattern)
        file_list = [
            f
            for f in glob.glob(kwargs["glob"], recursive=kwargs.get("glob_recursive", True))
            if Path(f).is_file()
        ]
    file_list += kwargs.get("files", [])
    if "file" in kwargs and kwargs["file"]:
        file_list.append(kwargs["file"])
    # Only return unique files that exist
    return [f for f in set(file_list) if os.path.isfile(f)]


def expand_io_sig(filename, config):
    """Expand io signatures.

    This expands the IO signatures in a block descriptor YAML file. After
    expansion, every entry in the IO signatures sub-dictionary is extended with
    detailed information about the wires used, their direction, etc.

    Arguments:
        filename: The name of the YAML file this data was pulled from.
        config: The contents of the block descriptor YAML file as a dictionary.

    Return value:
    The same config object, but the entries of the IO ports section has additional
    information about the wires.
    """
    from uhd import get_pkg_data_path

    signatures = yaml_utils.io_signatures(
        yaml_utils.get_core_config_path(get_pkg_data_path()), {filename: config}
    )
    config = yaml_utils.IOConfig(config, signatures)
    return config


class StepExecutor:
    """Main engine for executing steps in a command script."""

    class StepError(Exception):
        """Exception raised when a step fails."""

    def __init__(self, global_vars, args, cmd):
        """Initialize the executor."""
        assert isinstance(cmd, dict), "Command must be a dictionary"
        self.cmd = cmd
        self.args = args
        self.global_vars = global_vars
        self.template_base = os.path.join(os.path.dirname(__file__), "templates")
        self._log = logging.getLogger(__name__)

    def _setv(self, name, val):
        """Set variable name to val."""
        self.cmd["variables"][name] = val

    def _resolve(self, value, **extra_kwargs):
        """Shorthand for resolving a value."""
        return resolve(
            value,
            args=self.args,
            **self.global_vars,
            **self.cmd.get("variables", {}),
            os=os,
            **extra_kwargs,
        )

    def run(self, steps):
        """Run all the steps in the command."""
        for step in steps:
            for step_type, step_args in step.items():
                args_resolved = {}
                for k, v in step_args.items():
                    try:
                        args_resolved[k] = self._resolve(v)
                    except Exception as e:
                        raise StepExecutor.StepError(
                            f"Error resolving argument {k}={v} for step {step_type}: {str(e)}"
                        )
                args_resolved_visual = args_resolved.copy()
                if "steps" in args_resolved_visual:
                    args_resolved_visual["steps"] = str(len(args_resolved_visual["steps"]))
                self._log.debug(
                    "Executing step %s(%s)",
                    step_type,
                    ", ".join([f"{k}={v}" for k, v in args_resolved_visual.items()]),
                )
                try:
                    getattr(self, step_type)(**args_resolved)
                except StepExecutor.StepError as e:
                    self._log.error(str(e))
                    sys.exit(1)
                except Exception as e:
                    self._log.error("Unexpected error running step %s: %s", step_type, str(e))
                    self._log.debug(traceback.format_exc())
                    sys.exit(1)

    def run_if(self, condition, steps):
        """Run a block of steps if a condition is true."""
        if self._resolve(condition):
            self.run(steps)

    def run_subprocess(self, cmd, **kwargs):
        """Run a command as a subprocess.

        Arguments:
        - cmd: The command to run as a list of arguments.
        - dst_var: If provided, the output will be stored in this variable.
        """
        cmd = [self._resolve(arg) for arg in cmd]
        cmd = [arg for arg in cmd if arg]
        dst_var = kwargs.get("dst_var")
        try:
            self._log.debug("Running command: %s", cmd)
            proc_result = subprocess.run(
                cmd,
                check=True,
                env=os.environ.copy(),
                capture_output=dst_var is not None,
            )
        except subprocess.CalledProcessError as e:
            if "error_msg" in kwargs:
                self._log.error(kwargs["error_msg"])
            else:
                self._log.error("Command failed with return code %s", e.returncode)
            raise StepExecutor.StepError(f"Command failed: {e}")
        if dst_var:
            if proc_result.stdout:
                self._setv(dst_var, proc_result.stdout.decode("utf-8").strip())
            else:
                self._setv(dst_var, "")

    def copy_dir(self, src, dst, **kwargs):
        """Copy a directory from src to dest, recursively."""
        self._log.debug("Copying directory %s to %s", src, dst)
        ignore = None
        if "ignore_globs" in kwargs:

            def ignore_patterns(path, names):
                relpath = os.path.relpath(path, os.getcwd())
                return [
                    name
                    for name in names
                    if any(
                        fnmatch.fnmatch(os.path.normpath(os.path.join(relpath, name)), pattern)
                        for pattern in kwargs["ignore_globs"]
                    )
                ]

            ignore = ignore_patterns
        copytree_kwargs = {"ignore": ignore}
        if sys.version_info >= (3, 8):
            copytree_kwargs["dirs_exist_ok"] = True
        shutil.copytree(src, dst, **copytree_kwargs)

    def search_and_replace(self, **kwargs):
        """Search and replace text in a file.

        pattern: The pattern to search for
        repl: Replacement string
        count: Max number of replacements (default: 0, all)
        """
        file_list = get_file_list(**kwargs)
        for file in file_list:
            self._log.debug(
                "Editing file %s (replacing `%s' with `%s')",
                file,
                kwargs["pattern"],
                kwargs["repl"],
            )
            with open(file, "r", encoding="utf-8") as f:
                contents = f.read()
            count = kwargs.get("count", 0)
            contents, sub_count = re.subn(
                kwargs["pattern"],
                kwargs["repl"],
                contents,
                count=count,
                flags=re.MULTILINE | re.DOTALL,
            )
            if sub_count == 0 and not kwargs.get("quiet", False):
                self._log.warning("Pattern not found in file %s", os.path.abspath(file))
            with open(file, "w", encoding="utf-8") as f:
                f.write(contents)

    def chdir(self, dir):
        """Change the working directory."""
        os.chdir(os.path.normpath(dir))
        self.global_vars["CWD"] = os.getcwd()
        self._log.debug("Changed working directory to %s", os.getcwd())

    def multi_rename(self, pattern, repl, **kwargs):
        """Rename multiple files with a regex pattern."""
        paths_to_rename = glob.glob(kwargs["glob"], recursive=kwargs.get("glob_recursive", True))
        # make sure files are renamed before the directories which contain the files by
        # sorting and reversing the list, example:
        # - first rename file "old/old.cc" to "old/new.cc"
        # - then rename directory "old" to "new"
        paths_to_rename.sort(reverse=True)
        for path in paths_to_rename:
            head, tail = os.path.split(path)
            new_path = os.path.join(head, re.sub(pattern, repl, tail))
            if new_path != path:
                self._log.debug("Renaming %s to %s", path, new_path)
                shutil.move(path, new_path)

    def parse_descriptor(self, source, var="config", **kwargs):
        """Load a block descriptor file."""
        if not source:
            raise StepExecutor.StepError("Cannot parse: No descriptor file specified.")
        yaml = YAML(typ="safe", pure=True)
        self._log.debug("Loading descriptor file %s into variable %s", source, var)
        try:
            with open(source, "r", encoding="utf-8") as f:
                self.global_vars[var] = yaml.load(f)
            if kwargs.get("expand_io_sig"):
                self.global_vars[var] = expand_io_sig(source, self.global_vars[var])
        except FileNotFoundError:
            raise StepExecutor.StepError(f"Descriptor file {source} not found.")
        except Exception as e:
            raise StepExecutor.StepError(f"Error loading descriptor file {source}: {str(e)}")

    def write_template(self, template, dest, **kwargs):
        """Write a template file."""
        template_dir = os.path.join(
            self.template_base, kwargs.get("namespace", self.cmd.get("template_namespace", ""))
        )
        lookup = mako.lookup.TemplateLookup(directories=[template_dir])
        Path(dest).parent.mkdir(parents=True, exist_ok=True)
        tpl = Template(filename=os.path.join(template_dir, template), lookup=lookup)
        vars = self.cmd.get("variables", {}).copy()
        # Make sure standard template variables are available
        if "year" in kwargs:
            vars["year"] = kwargs["year"]
        elif "year" in vars:
            pass
        else:
            vars["year"] = datetime.datetime.now().year
        self._log.debug("Writing template %s to %s", template, dest)
        if os.path.exists(dest):
            self._log.warning("Overwriting existing file %s", dest)
        with open(dest, "w", encoding="utf-8") as f:
            f.write(tpl.render(**self.global_vars, **vars, render_wire_width=render_wire_width))

    def _insert_text(self, pattern, text, repl, **kwargs):
        """Insert text into a file based on a regex."""
        file_list = get_file_list(**kwargs)
        for file in file_list:
            self._log.debug("Editing file %s (inserting `%s')", file, text.strip())
            with open(file, "r", encoding="utf-8") as f:
                contents = f.read()
            count = kwargs.get("count", 1)
            contents, sub_count = re.subn(
                pattern, repl, contents, count=count, flags=re.MULTILINE | re.DOTALL
            )
            if sub_count == 0:
                self._log.warning("Pattern not found in file %s", file)
            with open(file, "w", encoding="utf-8") as f:
                f.write(contents)

    def insert_after(self, pattern, text, **kwargs):
        """Insert text after a pattern in a file."""
        self._insert_text(pattern, text, r"\g<0>" + text, **kwargs)

    def insert_before(self, pattern, text, **kwargs):
        """Insert text before a pattern in a file."""
        self._insert_text(pattern, text, text + r"\g<0>", **kwargs)

    def append(self, text, **kwargs):
        """Append text to a file."""
        file_list = get_file_list(**kwargs)
        for file in file_list:
            self._log.debug("Appending to file %s", file)
            with open(file, "a", encoding="utf-8") as f:
                f.write(text)

    def comment_out(self, character="#", **kwargs):
        """Modify all lines in range to prepend character.

        Note that the line range starts at 1, as it would with sed or in an
        editor.
        """
        line_range = kwargs.get("range", [1, -1])
        file_list = get_file_list(**kwargs)
        for file in file_list:
            self._log.debug("Commenting out lines in %s", file)
            with open(file, "r", encoding="utf-8") as f:
                contents = f.readlines()
            for line_no_minus_one, line in enumerate(contents):
                if line_no_minus_one + 1 >= line_range[0] and (
                    line_no_minus_one + 1 <= line_range[1] or line_range[1] == -1
                ):
                    contents[line_no_minus_one] = character + line
            with open(file, "w", encoding="utf-8") as f:
                f.writelines(contents)

    def rmtree(self, **kwargs):
        """Remove a directory tree."""
        dir_to_remove = kwargs["dir"]
        self._log.debug("Removing directory %s", dir_to_remove)
        if os.path.exists(dir_to_remove) and os.path.isdir(dir_to_remove):
            shutil.rmtree(dir_to_remove)

    def log(self, **kwargs):
        """Print a message to the log."""
        level = kwargs.get("level", "info")
        message = kwargs.get("msg", "")
        symbol = kwargs.get("symbol", "✅")
        if level == "info":
            message = f"{symbol}   {message}"
        getattr(self._log, level)(message)

    def find_file(self, dst_var, required=False, **kwargs):
        """Find a file matching a pattern and store its path in a variable.

        Arguments:
        - glob/file/files: The pattern to match files against. See get_file_list for details.
        - dst_var: The name of the variable to store the path in.
        - required: If true, then an error is raised if no file is found. If
                    false, the variable is set to None if no file is found. This
                    is the default.
        - error_msg: The message to display if no file is found and required is true.
        """
        file_list = get_file_list(**kwargs)
        if len(file_list) > 1:
            raise StepExecutor.StepError(f"More than one file found matching pattern.")
        if len(file_list) == 0:
            if required:
                raise StepExecutor.StepError(
                    kwargs.get("error_msg", "No file found matching pattern.")
                )
            self.cmd["variables"][dst_var] = None
        else:
            self.cmd["variables"][dst_var] = file_list[0]

    def find_executable(self, dst_var, name, **kwargs):
        """Find an executable in the PATH and store it in a variable.

        Parameters:
        - dst_var: The name of the variable to store the executable in.
        - name: The name of the executable to find.

        Optional:
        - error_msg: The message to display if the executable is not found.
        """
        if "error_msg" in kwargs:
            error_msg = kwargs["error_msg"]
        else:
            error_msg = f"Executable {name} not found in PATH."
        path = shutil.which(name)
        if path is None:
            raise StepExecutor.StepError(error_msg)
        self._log.debug("Found executable %s at %s", name, path)
        self._setv(dst_var, os.path.abspath(path))

    def exit(self, **kwargs):
        """Exit the script."""
        if "msg" in kwargs:
            self._log.info(kwargs["msg"])
        sys.exit(kwargs.get("code", 1))

    def input(self, dst_var, prompt, **kwargs):
        """Request input from the user.

        The input is captured by calling input(). Then, the following rules apply:
        - If nothing was entered, and a default value is provided, the default
          is used and this function returns.
        - If a type is provided, the input is converted to that type. This may
          fail, in which case the user is prompted again.
        - Next, if a check is provided, the input is checked against the
          predicate. If the check fails, the user is prompted again.
        - When the user is prompted to re-enter the input, we use the string
          from "check_msg" to provide additional information.
        """
        default = kwargs.get("default")
        prompt = prompt.strip()
        if default:
            prompt += f" [{default}]"
        prompt += " "
        val = input(prompt)

        def _filter_bool(s):
            """A 'smart' conversion from string to Boolean."""
            if isinstance(s, str) and s.lower() in ["y", "yes", "1", "true", "on"]:
                return True
            else:
                return False

        filters = {"bool": _filter_bool, "int": lambda s: int(s, 0), "float": float}
        if val == "" and "default" in kwargs:
            self._setv(dst_var, kwargs["default"])
            return
        elif "type" in kwargs:
            if kwargs["type"] in filters:
                try:
                    val = filters[kwargs["type"]](val.strip())
                except:
                    val = None
            else:
                raise StepExecutor.StepError(f"Unknown input type {kwargs['type']}")
        else:
            val = val.strip()

        def _check_input(val):
            """Check if the input is valid."""
            if val is None:
                return False
            if "check" not in kwargs:
                return True
            try:
                check_val = self._resolve("${ " + kwargs["check"] + "}", **{dst_var: val})
                if check_val:
                    return True
            except:
                pass
            return False

        if not _check_input(val):
            check_msg = kwargs.get("check_msg", "Invalid input!")
            print(check_msg)
            # Try again
            return self.input(dst_var, prompt, **kwargs)

        self._setv(dst_var, val)

    def fork(self, command, args, **kwargs):
        """Fork a command and wait for it to finish.

        The commands that can be forked are other rfnoc_modtool subcommands.
        """
        from .rfnoc_modtool import collect_commands, resolve_vars

        if not isinstance(args, Namespace):
            args = Namespace(**{k: self._resolve(v) for k, v in args.items()})
        cmds = collect_commands()
        if command not in cmds:
            raise StepExecutor.StepError(f"Unknown command {command}")
        cmd = cmds[command]
        for arg_name, arg_info in cmd.get("args", {}).items():
            if arg_name not in args:
                if "default" in arg_info:
                    setattr(args, arg_name, arg_info["default"])
                else:
                    raise StepExecutor.StepError(f"Missing required argument {arg_name}")
        assert not cmd.get(
            "skip_identify_module", False
        ), "Cannot fork a command that skips module identification"
        cmd = resolve_vars(cmd, self.global_vars, args)
        sub_executor = StepExecutor(self.global_vars.copy(), args, cmd)
        self._log.debug(
            "Forking command %s(%s)",
            command,
            ", ".join([f"{k}={v}" for k, v in args.__dict__.items()]),
        )
        sub_executor.run(cmds[command]["steps"])

    def set(self, name, value, is_global=False, **kwargs):
        """Set a variable in the command script."""
        self._log.debug("Setting variable %s to %s", name, value)
        if is_global:
            self.global_vars[name] = self._resolve(value)
        else:
            self._setv(name, self._resolve(value))



#
# Copyright 2023 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
"""
UHD Extension, helpers, and utilities for doing more with the available PL DRAM
"""

import time
from uhd import rfnoc
from uhd.usrp import StreamArgs
from uhd.types import TXMetadata, RXMetadata, RXMetadataErrorCode, StreamMode, StreamCMD, TimeSpec

def enumerate_radios(graph, radio_chans):
    """
    Return a list of radio block controllers/chan pairs to use for this test.
    """
    def unpack_rcp_spec(rcps):
        """
        Convert a radio/channel pair specification into a tuple of form
        (radio_block_id, radio_chan).

        Valid inputs (and their corresponding outputs are:
        - "0/Radio#0:0" -> ("0/Radio#0", 0)
        - "0/Radio#0" -> ("0/Radio#0", 0)      [Channel 0 is chosen as default)
        - ("0/Radio#0", 0) -> ("0/Radio#0", 0)
        """
        if isinstance(rcps, str):
            return (rcps.split(':', 2)[0], int(rcps.split(':', 2)[1])) \
                   if ':' in rcps else (rcps, 0)
        if isinstance(rcps, tuple) and len(rcps) == 2:
            return rcps
        raise RuntimeError(f"Unknown radio channel pair specification: {rcps}")
    radio_id_chan_pairs = [unpack_rcp_spec(r) for r in radio_chans]
    # Sanity checks
    available_radios = graph.find_blocks("Radio")
    radio_chan_pairs = []
    for rcp in radio_id_chan_pairs:
        if rcp[0] not in available_radios:
            raise RuntimeError(f"'{rcp[0]}' is not a valid radio block ID!")
        radio_chan_pairs.append((rfnoc.RadioControl(graph.get_block(rcp[0])), rcp[1]))
    return radio_chan_pairs

def find_replay_block(graph, replay_blockid):
    """
    Find any or a specific replay block
    """
    if replay_blockid is None:
        blocklist = graph.find_blocks("Replay")
        if not blocklist:
            raise RuntimeError("No Replay block found on RFNoC graph!")
        replay_blockid = blocklist[0]
    return rfnoc.ReplayBlockControl(graph.get_block(replay_blockid))

class DramTransmitter:
    """
    Helper class to stream data from DRAM to one or more radios.

    This can be useful when setting up a USRP as a transmitter, with a waveform
    preloaded into memory.

    NOTE: This assumes we are using a single memory bank, until UHD is upgraded
    to better handle multiple memory banks.

    Arguments:
    rfnoc_graph -- The graph object
    radio_chan -- A list of radio channels to connect to the DRAM, in one of
                  these formats:
                  - "0/Radio:0#0" (single string)
                  - "0/Radio:0" (single string, channel 0 by default)
                  - ("0/Radio#0", 0) (tuple of block id and channel number)
    replay_blockid -- If specified, use this replay block
    replay_ports -- List of replay ports to use. Must be at least as long as the 
                    list of radio channels. If not specified, use all ports.
                    Therefore this should be used if TX and RX are going to be used.
    cpu_format -- For the upload process, the data format to be used
    mem_regions -- A list of (memory start, memory size) tuples, one per replay block port.
                   If left out, the memory is split up evenly among available replay block ports.
    """
    def __init__(self,
                 rfnoc_graph,
                 radio_chans,
                 replay_blockid=None,
                 replay_ports=None,
                 cpu_format='fc32',
                 mem_regions=None,
                 ):
        # We make replay_blocks a list so we can support multiple replay blocks
        # (not only on multiple motherboards) in the future without changing APIs
        self.replay_blocks = [find_replay_block(rfnoc_graph, replay_blockid)]
        self.word_size = max(x.get_word_size() for x in self.replay_blocks)
        # In the radio, we always use sc16 regardless of cpu_format
        self.bytes_per_sample = 4
        if replay_ports is None:
            replay_ports = list(range(len(radio_chans)))
        self.replay_ports = replay_ports

        self.reconnect(rfnoc_graph, radio_chans, mem_regions)
        # Since for multi-channel we nevertheless only use one input port to upload the 
        # data we always take the first of the given replay ports
        self.replay_in_port = self.replay_ports[0]

        stream_args = StreamArgs(cpu_format, "sc16")
        self.tx_streamer = rfnoc_graph.create_tx_streamer(1, stream_args)

        rfnoc_graph.connect(
            self.tx_streamer, 0,
            self.replay_blocks[0].get_unique_id(), self.replay_in_port)
        rfnoc_graph.commit()

    def reconnect(self, rfnoc_graph, radio_chans, mem_regions=None):
        """
        Reconnect the replay block to a new set of radio channels.
        """
        self.replay_ports = self._sanitize_replay_ports(self.replay_ports)
        # Map requested radio_chans to the outputs of this replay block
        self.radio_chan_pairs = enumerate_radios(rfnoc_graph, radio_chans)
        if len(self.replay_ports) < len(self.radio_chan_pairs):
            raise RuntimeError(
                f"{len(self.radio_chan_pairs)} radio channels were requested to be "
                f"used with {len(self.replay_ports)} replay ports on "
                f"Replay block {self.replay_blocks[0].get_unique_id()}! "
                "There must be at least as many replay block ports as radio channels.")

        # Disconnect the ports of the replay block that we want to use later on
        active_conns = rfnoc_graph.enumerate_active_connections()
        for conn in active_conns:
            # If any of the connections that we require are taken already we will
            # disconnect them here. That is especially important if both the transmitter
            # and the receiver should be used.
            if (conn.src_blockid == self.replay_blocks[0].get_unique_id() and \
                conn.src_port in self.replay_ports) or \
               (conn.dst_blockid == self.replay_blocks[0].get_unique_id() and \
                conn.dst_port in self.replay_ports):
                if "Streamer" in conn.dst_blockid:
                    # Need to use streamer disconnect method for streamers
                    rfnoc_graph.disconnect(
                        conn.dst_blockid, conn.dst_port
                    )
                elif "Streamer" in conn.src_blockid:
                    pass
                else:
                    # Disconnect for all other kinds of blocks
                    rfnoc_graph.disconnect(
                        conn.src_blockid,
                        conn.src_port,
                        conn.dst_blockid,
                        conn.dst_port,
                    )

        self.duc_chan_pairs = []
        for replay_port_idx, rcp in enumerate(self.radio_chan_pairs):
            edge = rfnoc.connect_through_blocks(
                rfnoc_graph,
                self.replay_blocks[0].get_unique_id(), self.replay_ports[replay_port_idx],
                rcp[0].get_unique_id(), rcp[1], True)
            duc_edge = \
                next(filter(lambda edge: rfnoc.BlockID(edge.dst_blockid).match("DUC"), edge), None)
            if duc_edge:
                self.duc_chan_pairs.append((
                    rfnoc.DucBlockControl(rfnoc_graph.get_block(duc_edge.dst_blockid)),
                    duc_edge.dst_port
                ))
            else:
                self.duc_chan_pairs.append(None)
        # Assign memory regions to each output port (and leave space for potential input ports)
        if mem_regions is None:
            mem_stride = self.replay_blocks[0].get_mem_size() \
                // self.replay_blocks[0].get_num_output_ports()
            mem_stride -= mem_stride % self.word_size
            self.mem_regions = [
                (idx * mem_stride, mem_stride)
                for idx in range(self.replay_blocks[0].get_num_output_ports())
            ]
        else:
            # Create mem_regions list prepopulated with 0 and then fill it with the
            # given mem_regions. This way we can also handle the case where only
            # a subset of the available replay ports is used.
            self.mem_regions = self._sanitize_mem_regions(mem_regions)
        # This stores how much data we have currently uploaded to memory. We
        # initialize this with the full memory, not zero.
        self.upload_size = [x[1] for x in self.mem_regions]


    def _sanitize_replay_ports(self, ports):
        """
        Helper function to turn ports into a valid list of ports on the replay
        block.
        """
        if ports is None:
            ports = self.replay_ports
        if isinstance(ports, int):
            ports = [ports]
        if any((port >= self.replay_blocks[0].get_num_output_ports() for port in ports)):
            raise RuntimeError(
                    f"Invalid output port on replay block! Available ports: "
                    f"{self.replay_blocks[0].get_num_output_ports()} Requested: "
                    f"{ports}")
        return ports

    def _sanitize_mem_regions(self, custom_mem_regions=None):
        """
        Helper function to make a list of memory regions valid for this replay
        block and with given custom memory regions.
        """
        # If called without custom regions just return what we already have
        if custom_mem_regions is None:
            return self.mem_regions
        mem_regions = [(0,0)] * self.replay_blocks[0].get_num_output_ports()
        # If custom region is big enough for all possible ports, use it
        if len(mem_regions) == len(custom_mem_regions):
            return custom_mem_regions
        # Otherwise we have to distribute the given regions among the available ports
        for idx, mem_region in enumerate(custom_mem_regions):
            if idx < len(self.replay_ports):
                mem_regions[self.replay_ports[idx]] = mem_region
        return mem_regions

    def _upload(self, waveform, mem_start, mem_size):
        """
        Upload helper function

        Arguments:
        waveform: 1-dimensional numpy array with waveform data. Must be of the
                  same data type as specified during the constructor.
        mem_start: Memory address where the data should be uploaded to
        mem_size: Amount of available memory. If the data in waveforms exceeds
                  this value, then waveform will be only partially uploaded.
        """
        # Sanitize parameters
        assert mem_start < self.replay_blocks[0].get_mem_size(), \
            f"Invalid memory start location: {mem_start}"
        assert mem_start + mem_size <= self.replay_blocks[0].get_mem_size(), \
            f"Invalid memory range: {mem_start} - {mem_start + mem_size}"
        assert mem_start % self.word_size == 0, \
            f"Memory region start (0x{mem_start:X}) is not aligned with " \
            f"word size ({self.word_size})!"
        assert mem_size % self.word_size == 0, \
            f"Memory region size (0x{mem_size:X}) is not aligned with " \
            f"word size ({self.word_size})!"
        num_items = min(len(waveform), int(mem_size) // self.bytes_per_sample)
        waveform = waveform[:num_items]

        num_bytes = num_items * self.bytes_per_sample
        in_port = self.replay_in_port
        # Configure DRAM block for recording
        self.replay_blocks[0].record(mem_start, num_bytes, in_port)
        # Flush data on input buffer
        flush_timeout = time.monotonic() + .25
        while time.monotonic() < flush_timeout:
            if self.replay_blocks[0].get_record_fullness(in_port) == 0:
                break
            self.replay_blocks[0].record_restart(in_port)
        # Upload data
        tx_md = TXMetadata()
        tx_md.start_of_burst = True
        tx_md.end_of_burst = True
        if self.tx_streamer.send(waveform, tx_md, 10.0) != num_items:
            raise RuntimeError("Unable to upload all data without errors!")
        # Make sure DRAM is fully populated
        upload_timeout = time.monotonic() + 20.0
        while time.monotonic() < upload_timeout and \
                self.replay_blocks[0].get_record_fullness(in_port) < num_bytes:
            time.sleep(.05)
        fullness = self.replay_blocks[0].get_record_fullness(in_port)
        if fullness != num_bytes:
            raise RuntimeError(
                f"DRAM fullness did not reach expected levels! "
                f"{fullness}/{num_bytes} bytes.")
        return num_bytes

    def upload(self, waveform, ports=None, mem_regions=None):
        """
        Store a waveform to memory

        Arguments:
        ports: If this argument is given, then we use the mem_regions attribute
               of this class to identify where to store the waveform. If ports
               is a list, then the waveform will be uploaded once per port to
               the corresponding memory region.
        mem_regions: If this argument is given, ports is ignored. This will
                     directly specify the memory regions stored in this object.
                     NOTE: This class attempts to keep track of how many samples
                     have been uploaded into memory. If mem_regions is provided,
                     then the mapping between ports and memory regions is lost,
                     and this class cannot know how many samples are available
                     for a given port. Therefore, use this argument to
                     deliberately override where sample data should be stored,
                     e.g., to override sections of previously uploaded sample
                     data.

        If port is not specified, upload to all ports.
        """
        if mem_regions:
            ports = None
        else:
            ports = self._sanitize_replay_ports(ports)
            # Sanitize again in case self.mem_regions was changed from the outside
            mem_regions = self._sanitize_mem_regions(self.mem_regions)
            mem_regions = [ mem_regions[port] for port in ports]
        if len(mem_regions) == 2 and \
                isinstance(mem_regions, (tuple, list)) and \
                isinstance(mem_regions[0], int):
            mem_regions = [mem_regions]

        # If this method is called from send() then we will work on a per-channel
        # basis. Otherwise we will walk through the mem_regions that we have put
        # together above.
        if len(waveform.shape) == 1:
            waveform = [waveform] * len(mem_regions)
        if len(waveform) < len(mem_regions):
            raise RuntimeError("Number of waveforms in waveform array does not match "
                               "the number of memory regions!")
        for region_idx, mem_region in enumerate(mem_regions):
            # Since by default we slice the dram per input/output port, we only
            # want to upload to the memory regions that we are actually using.
            if ports is None or region_idx < len(ports):
                bytes_uploaded = self._upload(waveform[region_idx], *mem_region)
                if ports:
                    self.upload_size[ports[region_idx]] = bytes_uploaded

    def issue_stream_cmd(self, stream_cmd, ports=None):
        """
        Issue a command to start or stop the streaming from DRAM.

        If ports is not specified, issue the stream command on all ports.

        Under the hood, this will call replay_block_control::config_play() (if
        the stream command is a start-command) and replay_block_control::issue_stream_cmd().

        Note that this will configure the replay block for playback every time
        it is called. If upload() was called previously, then this class knows
        how many samples are available for a given port, and it will only use
        the occupied memory space. This means that if a subset of the available
        memory region for a port is occupied with valid sample data, and the
        number of samples requested in this stream command is either unlimited
        or larger than the available sample data, the replay block will correctly
        loop around the available sample data.
        """
        ports = self._sanitize_replay_ports(ports)
        mem_regions = self._sanitize_mem_regions(self.mem_regions)
        for port in ports:
            if stream_cmd.stream_mode in (
                    StreamMode.start_cont,
                    StreamMode.num_done,
                    StreamMode.num_more):
                mem_region = mem_regions[port]
                mem_size = min(self.upload_size[port], mem_region[1])
                self.replay_blocks[0].config_play(mem_region[0], mem_size, port)
            self.replay_blocks[0].issue_stream_cmd(stream_cmd, port)

    def send(self, data, metadata, timeout=0.1):
        """
        This is a wrapper around upload() and issue_stream_cmd() that can be
        used to use this class like you would use a TxStreamer object.

        Compared to calling those two functions directly, this is less flexible.
        It requires 'data' to be of the right shape (meaning that it needs as
        many sub-arrays as there are channels, just like when calling send() on
        a TxStreamer object), and it will only stream the data once (again, as
        if with a TX streamer object).

        Unlike TxStreamer.send(), you cannot call this repeatedly with small
        packets. Every call to this method will overwrite the previous call's
        data.

        The timeout parameter is unused, it is only there to retain the call
        signature compatibility. Time specs are pulled from the TX metadata object.
        """
        num_chans = len(self.radio_chan_pairs)
        num_samps = len(data) if len(data.shape) == 1 else data.shape[1]
        if num_samps == 0 or any((x == 0 for x in data.shape)):
            # When streaming to radio, we sometimes send an empty burst with an
            # EOB to receive the ACK. With the DRAM replay, this doesn't work so
            # well, and we always send an EOB from here anyway, so we just
            # pretend like we did this.
            return 0
        if (num_chans > 1 and len(data.shape) != 2) or (data.shape[0] < num_chans):
            raise RuntimeError(
                f"Number of TX channels {num_chans} does not match the dimensions "
                f"of the data array ({data.shape[0] if len(data.shape) == 2 else 1})")
        # First upload the data
        if num_chans == 1:
            if len(data.shape) == 2:
                data = data[0]
            self.upload(data, self.replay_ports[0])
        else:
            for chan, _ in enumerate(self.radio_chan_pairs):
                self.upload(data[chan], self.replay_ports[chan])
        # Then trigger stream command
        stream_cmd = StreamCMD(StreamMode.num_done)
        stream_cmd.stream_now = not metadata.has_time_spec
        stream_cmd.time_spec = metadata.time_spec
        stream_cmd.num_samps = num_samps
        # This defaults to "all ports"
        self.issue_stream_cmd(stream_cmd)
        return num_samps

    def recv_async_msg(self, timeout=0.1):
        """
        This emulates TxStreamer.recv_async_msg().
        """
        return self.replay_blocks[0].get_play_async_metadata(timeout)

class DramReceiver:
    """
    Helper class to stream data from one or more radios to DRAM.

    This can be useful when setting up a USRP as a receiver.

    NOTE: This assumes we are using a single memory bank, until UHD is upgraded
    to better handle multiple memory banks.

    Arguments:
    rfnoc_graph -- The graph object
    radio_chans -- A list of radio channels to connect to the DRAM, in one of
                  these formats:
                  - "0/Radio:0#0" (single string)
                  - "0/Radio:0" (single string, channel 0 by default)
                  - ("0/Radio#0", 0) (tuple of block id and channel number)
    replay_blockid -- If specified, use this replay block
    replay_ports -- List of replay ports to use. Must be at least as long as the 
                    list of radio channels. If not specified, use all ports.
                    Therefore this should be used if TX and RX are going to be used.
    cpu_format -- Desired data format of the downloaded, received data on the host.
    mem_regions -- A list of (memory start, memory size) tuples, one per replay block port.
                   If left out, the memory is split up evenly among available replay block ports.
    throttle -- Throttle factor for the streamer. This is a value between 0 and
                1 or a percentage in the range (0%, 100%] that is passed as string.
    """
    def __init__(self,
                 rfnoc_graph,
                 radio_chans,
                 replay_blockid=None,
                 replay_ports=None,
                 cpu_format='fc32',
                 mem_regions=None,
                 throttle="0.1"
                 ):
        # We make replay_blocks a list so we can support multiple replay blocks
        # (not only on multiple motherboards) in the future without changing APIs
        self.replay_blocks = [find_replay_block(rfnoc_graph, replay_blockid)]
        self.word_size = max(x.get_word_size() for x in self.replay_blocks)
        # In the radio, we always use sc16 regardless of cpu_format
        self.bytes_per_sample = 4
        stream_args = StreamArgs(cpu_format, "sc16")
        stream_args.args['throttle'] = throttle
        if replay_ports is None:
            replay_ports = list(range(len(radio_chans)))
        self.replay_ports = replay_ports
        self.receive_metadata = None

        self.reconnect(rfnoc_graph, radio_chans, mem_regions)
        # We only use the first of the given replay ports to download the data sequentially.
        self.replay_out_port = self.replay_ports[0]

        self.rx_streamer = rfnoc_graph.create_rx_streamer(1, stream_args)

        rfnoc_graph.connect(
            self.replay_blocks[0].get_unique_id(), self.replay_out_port,
            self.rx_streamer, 0)
        rfnoc_graph.commit()

    def reconnect(self, rfnoc_graph, radio_chans, mem_regions=None):
        """
        Reconnect the replay block to a new set of radio channels.
        """
        self.replay_ports = self._sanitize_replay_ports(self.replay_ports)
        # Map requested radio_chans to the inputs of this replay block
        self.radio_chan_pairs = enumerate_radios(rfnoc_graph, radio_chans)
        if len(self.replay_ports) < len(self.radio_chan_pairs):
            raise RuntimeError(
                f"{len(self.radio_chan_pairs)} radio channels were requested to be "
                f"used with {len(self.replay_ports)} replay ports on "
                f"Replay block {self.replay_blocks[0].get_unique_id()}!"
                "There must be at least as many replay block ports as radio channels.")

        # Disconnect the ports of the replay block that we want to use later on
        active_conns = rfnoc_graph.enumerate_active_connections()
        for conn in active_conns:
            # If any of the connections that we require are taken already we will
            # disconnect them here. That is especially important if both the transmitter
            # and the receiver should be used.
            if (conn.src_blockid == self.replay_blocks[0].get_unique_id() and \
                conn.src_port in self.replay_ports) or \
               (conn.dst_blockid == self.replay_blocks[0].get_unique_id() and \
                conn.dst_port in self.replay_ports):
                if "Streamer" in conn.src_blockid:
                    # Need to use streamer disconnect method for streamers
                    rfnoc_graph.disconnect(
                        conn.dst_blockid, conn.dst_port
                    )
                elif "Streamer" in conn.dst_blockid:
                    pass
                else:
                    # Disconnect for all other blocks
                    rfnoc_graph.disconnect(
                        conn.src_blockid,
                        conn.src_port,
                        conn.dst_blockid,
                        conn.dst_port,
                    )

        self.ddc_chan_pairs = []
        for replay_port_idx, rcp in enumerate(self.radio_chan_pairs):
            edge = rfnoc.connect_through_blocks(
                rfnoc_graph,
                rcp[0].get_unique_id(), rcp[1],
                self.replay_blocks[0].get_unique_id(), self.replay_ports[replay_port_idx], True)
            ddc_edge = \
                next(filter(lambda edge: rfnoc.BlockID(edge.dst_blockid).match("DDC"), edge), None)
            if ddc_edge:
                self.ddc_chan_pairs.append((
                    rfnoc.DdcBlockControl(rfnoc_graph.get_block(ddc_edge.dst_blockid)),
                    ddc_edge.dst_port
                ))
            else:
                self.ddc_chan_pairs.append(None)

        # Assign memory regions to each input port (and leave space for potential output ports)
        if mem_regions is None:
            mem_stride = self.replay_blocks[0].get_mem_size() \
                // self.replay_blocks[0].get_num_input_ports()
            mem_stride -= mem_stride % self.word_size
            self.mem_regions = [
                (idx * mem_stride, mem_stride)
                for idx in range(self.replay_blocks[0].get_num_input_ports())
            ]
        else:
            self.mem_regions = self._sanitize_mem_regions(mem_regions)
        # This stores how much data we have currently downloaded to memory. We
        # initialize this with the full memory, not zero.
        self.download_size = [x[1] for x in self.mem_regions]


    def _sanitize_replay_ports(self, ports):
        """
        Helper function to turn ports into a valid list of ports on the replay
        block.
        """
        if ports is None:
            ports = self.replay_ports
        if isinstance(ports, int):
            ports = [ports]
        if any((port >= self.replay_blocks[0].get_num_input_ports() for port in ports)):
            raise RuntimeError(
                    f"Invalid input port on replay block! Available ports: "
                    f"{self.replay_blocks[0].get_num_input_ports()} Requested: "
                    f"{ports}")
        return ports

    def _sanitize_mem_regions(self, custom_mem_regions=None):
        """
        Helper function to make a list of memory regions valid for this replay
        block and with given custom memory regions.
        """
        # If called without custom regions just return what we already have
        if custom_mem_regions is None:
            return self.mem_regions
        mem_regions = [(0,0)] * self.replay_blocks[0].get_num_input_ports()
        # If custom region is big enough for all possible ports, use it
        if len(mem_regions) == len(custom_mem_regions):
            return custom_mem_regions
        # Otherwise we have to distribute the given regions among the available ports
        for idx, mem_region in enumerate(custom_mem_regions):
            if idx < len(self.replay_ports):
                mem_regions[self.replay_ports[idx]] = mem_region
        return mem_regions

    def _download(self, waveform, mem_start, mem_size):
        """
        Download helper function

        Arguments:
        waveform: 1-dimensional numpy array with waveform data. Must be of the
                  same data type as specified during the constructor.
        mem_start: Memory address where the data should be downloaded from
        mem_size: Amount of available memory. This is the maximum length that 
                  a captured waveform can have.
        """
        # Sanitize parameters
        assert mem_start < self.replay_blocks[0].get_mem_size(), \
            f"Invalid memory start location: {mem_start}"
        assert mem_start + mem_size <= self.replay_blocks[0].get_mem_size(), \
            f"Invalid memory range: {mem_start} - {mem_start + mem_size}"
        assert mem_start % self.word_size == 0, \
            f"Memory region start (0x{mem_start:X}) is not aligned with " \
            f"word size ({self.word_size})!"
        assert mem_size % self.word_size == 0, \
            f"Memory region size (0x{mem_size:X}) is not aligned with " \
            f"word size ({self.word_size})!"
        num_items = min(len(waveform), int(mem_size) // self.bytes_per_sample)
        num_bytes = num_items * self.bytes_per_sample
        out_port = self.replay_out_port
        stream_cmd = StreamCMD(StreamMode.num_done)
        stream_cmd.num_samps = num_items
        stream_cmd.time_spec = TimeSpec(0.0)
        self.replay_blocks[0].config_play(mem_start, num_bytes, out_port)
        self.rx_streamer.issue_stream_cmd(stream_cmd)

        if not self.receive_metadata:
            self.receive_metadata = RXMetadata()
        num_items = self.rx_streamer.recv(waveform, self.receive_metadata, 15.0)
        if self.receive_metadata.error_code != RXMetadataErrorCode.none:
            # While the error code might be overwritten by the next call to _download(),
            # returning 0 will lead to recv() returning 0, too, which indicates an error.
            return 0
        return num_items

    def download(self, waveform, ports=None, mem_regions=None):
        """
        Download a waveform from memory to host

        Arguments:
        ports: If this argument is given, then we use the mem_regions attribute
               of this class to identify where the waveform is stored. If ports
               is a list, then the waveform will be downloaded once per port from
               the corresponding memory region.
        mem_regions: If this argument is given, ports is ignored. This will
                     directly specify the memory regions stored in this object.
                     NOTE: This class attempts to keep track of how many samples
                     have been sampled into memory. If mem_regions is provided,
                     then the mapping between ports and memory regions is lost,
                     and this class cannot know how many samples are available
                     for a given port. Therefore, use this argument to
                     deliberately override where sample data should be read from.

        If port is not specified, download from all ports.
        """
        if mem_regions:
            ports = None
        else:
            ports = self._sanitize_replay_ports(ports)
            # Sanitize again in case self.mem_regions was changed from the outside
            mem_regions = self._sanitize_mem_regions(self.mem_regions)
            mem_regions = [ mem_regions[port] for port in ports]
        if len(mem_regions) == 2 and \
                isinstance(mem_regions, (tuple, list)) and \
                isinstance(mem_regions[0], int):
            mem_regions = [mem_regions]

        # If this method is called from recv() then we will work on a per-channel
        # basis. Otherwise we will walk through the mem_regions that we have put
        # together above.
        if len(waveform.shape) == 1:
            bytes_downloaded = self._download(waveform, *mem_regions[0])
            self.download_size[ports[0]] = bytes_downloaded
        else:
            for region_idx, mem_region in enumerate(mem_regions):
                if ports is None or region_idx < len(self.radio_chan_pairs):
                    bytes_downloaded = self._download(waveform[region_idx], *mem_region)
                    if ports:
                        self.download_size[ports[region_idx]] = bytes_downloaded

    def issue_stream_cmd(self, stream_cmd, ports=None):
        """
        Issue a command to start or stop the streaming to DRAM.

        If ports is not specified, issue the stream command on all ports.
        """
        assert stream_cmd.stream_mode == StreamMode.num_done, \
            f"Invalid stream mode: {stream_cmd.stream_mode}"
        ports = self._sanitize_replay_ports(ports)
        mem_regions = self._sanitize_mem_regions(self.mem_regions)
        # Create a copy of the pointer to the original stream command to be able to edit it in case
        # we have to adjust the number of samples that the radio will send.
        tmp_stream_cmd = stream_cmd
        for idx, rcp in enumerate(self.radio_chan_pairs):
            stream_cmd = tmp_stream_cmd
            # Flush data on output buffer
            flush_timeout = time.monotonic() + .25
            while time.monotonic() < flush_timeout:
                if self.replay_blocks[0].get_record_fullness(ports[idx]) == 0:
                    break
                self.replay_blocks[0].record_restart(ports[idx])
            mem_region = mem_regions[ports[idx]]
            mem_size = min(stream_cmd.num_samps * self.bytes_per_sample, mem_region[1])
            self.replay_blocks[0].record(mem_region[0], mem_size, ports[idx])
            # In case we're using a DDC, we need to adjust the number of samples that the radio
            # will send, so that after down-converting it meets what the replay block expects.
            if self.ddc_chan_pairs[idx]:
                rate_ratio = self.ddc_chan_pairs[idx][0].get_input_rate(
                    self.ddc_chan_pairs[idx][1]) / \
                    self.ddc_chan_pairs[idx][0].get_output_rate(
                    self.ddc_chan_pairs[idx][1])
                stream_cmd = StreamCMD(StreamMode.num_done)
                stream_cmd.num_samps = int(tmp_stream_cmd.num_samps * rate_ratio)
                stream_cmd.stream_now = tmp_stream_cmd.stream_now
                stream_cmd.time_spec = tmp_stream_cmd.time_spec
            rcp[0].issue_stream_cmd(stream_cmd, rcp[1])

        timeout = time.monotonic() + 15.0
        while any((self.replay_blocks[0].get_record_fullness(ports[idx]) < mem_size
            for idx, _ in enumerate(self.radio_chan_pairs))):
            time.sleep(0.200)
            if time.monotonic() > timeout:
                raise RuntimeError("Timeout while loading replay buffer!")

    def recv(self, data, metadata, timeout=0.1):
        """
        This is a wrapper around download() that can be used to use this class
        like you would use an RxStreamer object.

        Unlike RxStreamer.recv(), you cannot call this repeatedly with small
        packets. Every call to this method will overwrite the previous call's
        data.

        Unlike RxStreamer.recv(), DramReceiver.recv() streams the received data
        from the replay block to the host on a per-channel basis. Therefore it
        generates RxMetadata for each channel which is not merged into a single
        metadata object. Instead, the last received metadata object is stored.
        If no error occured this should be the same for all channels. If an error
        occurs, the _download() method will throw an error which differs from the
        behavior of the RxStreamer.recv() method.

        The timeout parameter is unused, it is only there to retain the call
        signature compatibility. Time specs are pulled from the RX metadata object.
        """
        num_chans = len(self.radio_chan_pairs)
        self.receive_metadata = metadata
        if num_chans == 1:
            self.download(data, self.replay_ports[0])
        else:
            for idx, _ in enumerate(self.radio_chan_pairs):
                self.download(data[idx], self.replay_ports[idx])
        return min(self.download_size)



#
# Copyright 2024 Ettus Research, a National Instruments Brand
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
""" @package usrp_clock
Python UHD module containing the MultiUSRPClock
"""

from .. import libpyuhd as lib


class MultiUSRPClock(lib.usrp_clock.multi_usrp_clock):
    """
    MultiUSRPClock object for controlling devices
    """
    def __init__(self, args=""):
        """MultiUSRPClock constructor"""
        super(MultiUSRPClock, self).__init__(args)




"""Image builder: Build dispatch.

This is the entrypoint into the module from rfnoc_image_builder.py.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later

RFNoC image builder: All the algorithms required to turn either a YAML
description or a GRC file into an rfnoc_image_core.v file.
"""

import glob
import logging
import os
import pathlib
import re
import shutil
import subprocess
import sys

import mako.lookup
from mako import exceptions

from .. import get_pkg_data_path
from . import grc, yaml_utils
from .builder_config import ImageBuilderConfig
from .template import Template
from .utils import check_include_paths_backward_compat

### DATA ######################################################################
# Directory under the FPGA repo where the device directories are

USRP3_LIB_RFNOC_DIR = os.path.join("usrp3", "lib", "rfnoc")

# Path to the system's bash executable
BASH_EXECUTABLE = "/bin/bash"  # FIXME this should come from somewhere

# Map device names to the corresponding directory under usrp3/top
DEVICE_DIR_MAP = {
    "x300": "x300",
    "x310": "x300",
    "e300": "e300",
    "e310": "e31x",
    "e320": "e320",
    "n300": "n3xx",
    "n310": "n3xx",
    "n320": "n3xx",
    "x400": "x400",
    "x410": "x400",
    "x440": "x400",
}

# Picks the default make target per device
DEVICE_DEFAULTTARGET_MAP = {
    "x300": "X300_HG",
    "x310": "X310_HG",
    "x410": "X410",
    "x440": "X440",
    "e310": "E310_SG3",
    "e320": "E320_1G",
    "n300": "N300_HG",
    "n310": "N310_HG",
    "n320": "N320_XG",
}

# Secure core RTL hierarchical path
SECURE_CORE_INST_PATH = {
    "x300": "x300_core/bus_int/rfnoc_sandbox_i/secure_image_core_i",
    "x310": "x300_core/bus_int/rfnoc_sandbox_i/secure_image_core_i",
    "e310": "e31x_core_inst/rfnoc_image_core_i/secure_image_core_i",
    "e320": "e320_core_i/rfnoc_sandbox_i/secure_image_core_i",
    "n300": "n3xx_core/rfnoc_sandbox_i/secure_image_core_i",
    "n310": "n3xx_core/rfnoc_sandbox_i/secure_image_core_i",
    "n320": "n3xx_core/rfnoc_sandbox_i/secure_image_core_i",
    "x400": "x4xx_core_i/rfnoc_image_core_i/secure_image_core_i",
    "x410": "x4xx_core_i/rfnoc_image_core_i/secure_image_core_i",
    "x440": "x4xx_core_i/rfnoc_image_core_i/secure_image_core_i",
}


def get_vivado_path(fpga_top_dir, args):
    """Return path to Vivado installation directory."""
    viv_path = None
    if args.get("vivado_path"):
        viv_path = args["vivado_path"]
    else:
        get_viv_path_cmd = '. ./setupenv.sh && echo "VIVADO_PATH=$VIVADO_PATH"'
        try:
            output = subprocess.check_output(
                f'{BASH_EXECUTABLE} -c "{get_viv_path_cmd}"',
                shell=True,
                cwd=fpga_top_dir,
                text=True,
            )
            viv_path = re.search(r"VIVADO_PATH=(.*)", output).group(1)
        except subprocess.CalledProcessError:
            logging.error("Failed to get Vivado path from setupenv.sh")
            sys.exit(1)
    logging.debug("Using Vivado path: %s", viv_path)
    # check if viv_path is a directory
    if not os.path.isdir(viv_path):
        logging.error("Vivado path not found: %s", viv_path)
        sys.exit(1)
    return viv_path


def write_verilog(config, destination, args, template):
    """Generate Verilog output from a template.

    Mako templates from local template folder are used to generate the image
    core file. The template engine does not do any computation on the script
    parameter. Instead all necessary dependencies are resolved in this script
    to enforce early failure which is easier to track than errors in the
    template engine.
    :param config: ImageBuilderConfig derived from script parameter
    :param destination: Filepath to write to
    :param args: Dictionary of arguments for the code generation
    :param template: Mako template to use
    :return: None
    """
    template_dir = os.path.join(os.path.dirname(__file__), "templates")
    lookup = mako.lookup.TemplateLookup(directories=[template_dir])
    tpl_filename = os.path.join(template_dir, template)
    tpl = Template(filename=tpl_filename, lookup=lookup)
    try:
        block = tpl.render(**{"config": config, "args": args})
    except:
        print(exceptions.text_error_template().render())
        sys.exit(1)

    logging.debug("Writing output to %s", destination)
    with open(destination, "w", encoding="utf-8") as image_core_file:
        image_core_file.write(block)


def patch_netlist_constraints(device, build_dir):
    """Update constraints for the secure_image_core netlist.

    Updates the constraints for the secure_image_core netlist so that they will
    work when applied to the top-level design. The constraints output by Vivado
    assume the netlist is the top level. We need to adjust the constraints so
    that they apply to the correct level of the hierarchy when applied to the
    actual top level.

    :param device: The device to build for (x310, x410, e320, etc).
    :param build_dir: The build directory where all the build artifacts are stored.
    :return: None
    """
    if device not in SECURE_CORE_INST_PATH:
        logging.warning(f"Device {device} not found. Skipping secure core netlist patch.")
        return
    netlist_root = SECURE_CORE_INST_PATH[device]
    # Make a backup of "secure_image_core.xdc" in the artifact directory
    secure_xdc = os.path.join(build_dir, "secure_image_core.xdc")
    secure_xdc_bak = os.path.join(build_dir, "secure_image_core.xdc.bak")
    shutil.copy(secure_xdc, secure_xdc_bak)
    # Update all calls to "current_instance" to have the correct relative path.
    # Some examples:
    # current_instance                  -> current_instance {new_path}
    # current_instance -quiet           -> current_instance -quiet {new_path}
    # current_instance {path/in/design} -> current_instance {new_path/path/in/design}
    with open(secure_xdc, "r") as file:
        constraints = file.read()
    # Update path in calls that already include a path
    constraints = re.sub(
        r"^(\s*current_instance\s*(?:-quiet)?)\s*{(.*)}\s*$",
        rf"\1 {{{netlist_root}/\2}}",
        constraints,
        flags=re.MULTILINE,
    )
    # Substitute calls that don't include a path
    constraints = re.sub(
        r"^(\s*current_instance\s*(?:-quiet)?)\s*$",
        rf"\1 {{{netlist_root}}}",
        constraints,
        flags=re.MULTILINE,
    )
    # Set the current instance at the start and reset it at the end
    constraints = (
        f"current_instance {{{netlist_root}}}\n" + constraints + "\ncurrent_instance -quiet\n"
    )
    # Replace the constraints file with the new constraints
    with open(secure_xdc, "w") as file:
        file.write(constraints)


def gen_make_command(
    args, build_dir, build_output_dir, build_ip_dir, device, use_secure_netlist, makefile_src_paths
):
    """Generate the 'make' command that will build the desired bitfiles.

    This generates a full make command, including all the necessary options.
    """
    target = args["target"] if "target" in args else ""
    return (
        f"make {default_target(device, target)} "
        f"BUILD_DIR={build_dir} IMAGE_CORE_NAME={args['image_core_name']}"
        + (
            f" SECURE_CORE=1 SECURE_KEY={args.get('secure_key_file')}"
            if args.get("secure_core")
            else ""
        )
        + (" SECURE_NETLIST=1" if use_secure_netlist else "")
        + (" GUI=1 " if args.get("GUI") else "")
        + (" CHECK=1" if args.get("check_hdl") else "")
        + (" SYNTH=1" if args.get("synthesize_only") else "")
        + (
            " RFNOC_OOT_MAKEFILE_SRCS=" + "\\ ".join(makefile_src_paths)
            if makefile_src_paths
            else ""
        )
        + (" --jobs " + args["num_jobs"] if args.get("num_jobs") else "")
        + (" PROJECT=1" if args.get("save_project") else "")
        + (" IP_ONLY=1" if args.get("ip_only") else "")
        + " BUILD_OUTPUT_DIR="
        + build_output_dir
        + " BUILD_IP_DIR="
        + build_ip_dir
    )


def build(fpga_top_dir, device, build_dir, use_secure_netlist, base_dir, **args):
    """Call FPGA toolchain to actually build the image.

    :param fpga_top_dir: A path to the FPGA device source directory
                         (/path/to/uhd/fpga/usrp3/top/<device>)
    :param device: The device to build for (x310, x410, e320, etc). The name must
                   match the top-level directory under usrp3/top.
    :param build_make_cmd: The make command to actually build the bitfile. Targets
                           like 'cleanall' and so on are all appended by this
                           function.
    :param build_dir: The build directory where all the build artifacts are stored.
                         (e.g., .../usrp3/x400/build-x410_XG_200_rfnoc_image_core/)
    :param use_secure_netlist: Boolean for whether the build with secure image
                               core netlist
    :param base_dir: The base directory
    :param **args: Additional options
                   target: The target to build (leave empty for default).
                   clean_all: passed to Makefile
                   GUI: passed to Makefile
                   save_project: passed to Makefile
                   ip_only: passed to Makefile
                   num_jobs: Number of make jobs to use
                   build_output_dir: Custom bitstream output directory
                   source: The source of the build (YAML or GRC file path)
                   include_paths: List of paths to OOT modules
                   that don't follow the OOT module layout. These paths must
                   point directly to a Makefile.srcs file.
    :return: exit value of build process
    """
    ret_val = 0
    build_dir = os.path.abspath(build_dir)
    if not os.path.isdir(fpga_top_dir):
        logging.error("Not a valid directory: %s", fpga_top_dir)
        return 1
    build_output_dir = os.path.abspath(
        args.get("build_output_dir") or os.path.join(base_dir, "build")
    )
    build_ip_dir = os.path.abspath(args.get("build_ip_dir") or os.path.join(base_dir, "build-ip"))
    makefile_src_paths = [
        os.path.join(os.path.abspath(os.path.normpath(x)), os.path.join("fpga", "Makefile.srcs"))
        for x in args.get("include_paths", [])
    ] + args.get("extra_makefile_srcs", [])
    setup_cmd = ". ./setupenv.sh "
    if "vivado_path" in args and args["vivado_path"]:
        setup_cmd += "--vivado-path=" + args["vivado_path"] + " "
    make_cmd = ""
    if "clean_all" in args and args["clean_all"]:
        make_cmd += "make cleanall && "
    make_cmd += gen_make_command(
        args,
        build_dir,
        build_output_dir,
        build_ip_dir,
        device,
        use_secure_netlist,
        makefile_src_paths,
    )

    if args.get("generate_only"):
        logging.info("Skipping build (--generate-only option given)!")
        logging.info("Use the following command to manually build the image: %s", make_cmd)
        return 0

    make_cmd = setup_cmd + "&& " + make_cmd
    logging.info("Launching build with the following settings:")
    logging.info(" * FPGA Directory: %s", fpga_top_dir)
    logging.info(" * Build Artifacts Directory: %s", build_dir)
    logging.info(" * Build Output Directory: %s", build_output_dir)
    logging.info(" * Build IP Directory: %s", build_ip_dir)
    # Wrap it into a bash call:
    make_cmd = f'{BASH_EXECUTABLE} -c "{make_cmd}"'
    logging.info("Executing the following command: %s", make_cmd)
    my_env = os.environ.copy()
    ret_val = subprocess.call(make_cmd, shell=True, env=my_env, cwd=fpga_top_dir)
    if ret_val == 0 and args.get("secure_core"):
        patch_netlist_constraints(device, build_dir)
    logging.info("Build finished with return code %d.", ret_val)
    logging.info("It was launched with the following settings:")
    logging.info(" * FPGA Directory: %s", fpga_top_dir)
    logging.info(" * Build Artifacts Directory: %s", build_dir)
    logging.info(" * Build Output Directory: %s", build_output_dir)
    logging.info(" * Build IP Directory: %s", build_ip_dir)
    return ret_val


def target_dir(device):
    """Return target directory derived from chosen device.

    :param device: device to build for
    :return: target directory (relative path)
    """
    if not device.lower() in DEVICE_DIR_MAP:
        logging.error(
            "Unsupported device %s. Supported devices are %s", device, DEVICE_DIR_MAP.keys()
        )
        sys.exit(1)
    return DEVICE_DIR_MAP[device.lower()]


def default_target(device, target):
    """Select a valid target.

    If no target specified, selects the default building target based on the
    targeted device
    """
    if target is None:
        return DEVICE_DEFAULTTARGET_MAP.get(device.lower())
    return target


def load_module_yamls(include_paths):
    """Load all known descriptor YAMLs."""

    def load_module_descs(module_type):
        """Load separate block/module defs."""
        paths = [
            os.path.abspath(os.path.normpath(os.path.join(x, module_type))) for x in include_paths
        ]
        logging.debug("Looking for %s descriptors in:", module_type[:-1])
        for path in paths:
            logging.debug("    %s", os.path.normpath(path))
        return yaml_utils.read_yaml_definitions(*paths)

    known_modules = {
        module_type: load_module_descs(module_type)
        for module_type in ("blocks", "modules", "transport_adapters", "includes")
    }
    # In the path structure, blocks are in blocks/, but the image builder wants
    # this to be called noc_blocks to match the YAML.
    known_modules["noc_blocks"] = known_modules.pop("blocks")
    return known_modules


def load_image_core_source(
    source,
    source_type,
    config_path,
    core_config_path,
    device,
    target,
    image_core_name,
    known_modules,
):
    """Load image configuration.

    The configuration can be either passed as RFNoC image configuration or as
    GNU Radio Companion grc. In latter case the grc files is converted into a
    RFNoC image configuration on the fly.
    :param source: Path to the source f ile
    :param source_type: Type of the source file (yaml or grc)
    :return: image configuration as dictionary
    """

    def get_image_core_name(config):
        icore_name = image_core_name or config.get("image_core_name")
        if not icore_name:
            icore_name = os.path.splitext(os.path.basename(source))[0]
        return icore_name

    if source_type == "yaml":
        config = yaml_utils.load_config_validate(source, config_path, True)
    # Otherwise, it's a GRC file
    elif source_type == "grc":
        config = yaml_utils.read_yaml(source)
        logging.info("Converting GNU Radio Companion file to image builder format")
        config = grc.convert_to_image_config(config, core_config_path, known_modules)
        # Run it through ruamel.yaml backwards and forwards to guarantee the
        # same object types as if we had loaded it directly from a YAML file.
        config = yaml_utils.reload_dict(config)
    else:
        logging.error("No configuration file provided")
        sys.exit(1)
    device = device or config.get("device")
    target = target or config.get("default_target")
    return config, device, get_image_core_name(config), target


def build_image(repo_fpga_path, device, **args):
    """Generate image dependent Verilog code and run FPGA toolchain, if requested.

    :param repo_fpga_path: A path that holds the FPGA sources (/path/to/uhd/fpga).
                           Under this path, there should be a usrp3/top/
                           directory.
    :param device: The device to build for.
    :param **args: Additional options including
                   target: The target to build (leave empty for default).
                   generate_only: Do not build the bitfile after generating the code.
                   clean_all: passed to Makefile
                   GUI: passed to Makefile
                   save_project: passed to Makefile
                   ip_only: passed to Makefile
                   build_output_dir: passed to Makefile
                   include_paths: Paths to additional blocks
                   image_core_name: The name of the image core to be generated
                                    (e.g., usrp_x410_fpga_CG_400)
    :return: Exit result of build process or 0 if generate-only is given.
    """
    config_path = get_pkg_data_path()
    # We start by loading some core/standard files that are always needed.
    core_config_path = yaml_utils.get_core_config_path(config_path)
    include_paths = check_include_paths_backward_compat(args.get("include_paths", [])) + [
        os.path.join(config_path, "rfnoc")
    ]
    logging.debug("Include paths: %s", ";".join(include_paths))
    # A list of all known module descriptors
    known_modules = load_module_yamls(include_paths)
    # resolve signature after modules have been loaded (the module YAML files
    # may contain signatures themselves)
    signatures_conf = yaml_utils.io_signatures(core_config_path, *list(known_modules.values()))

    # Next, load the image core configuration
    assert "source" in args
    config, device, image_core_name, target = load_image_core_source(
        args["source"],
        args["source_type"],
        config_path,
        core_config_path,
        device,
        args.get("target"),
        args.get("image_core_name"),
        known_modules,
    )
    # If no target specified on the command line, use the default target from the
    # yaml configuration file.
    if args.get("target") is None and target is not None:
        args["target"] = target
    logging.info("Selected device: %s", device)
    assert image_core_name
    args["image_core_name"] = image_core_name
    logging.debug("Image core name: %s", image_core_name)
    if "base_dir" in args and args["base_dir"]:
        base_dir = args["base_dir"]
    else:
        base_dir = os.getcwd()
    args.pop("base_dir", None)
    logging.debug("Using base directory: %s", base_dir)
    if "build_dir" in args and args["build_dir"]:
        build_dir = args["build_dir"]
    else:
        build_dir = os.path.join(base_dir, "build-" + image_core_name)
    args.pop("build_dir", None)
    logging.debug("Using build artifacts directory: %s", build_dir)
    if pathlib.Path(build_dir).is_file():
        logging.error("Build artifacts directory is pointing to a file: %s", build_dir)
        return 1
    elif pathlib.Path(build_dir).is_dir():
        logging.info("Build artifacts directory already exists (contents will be overwritten).")
    else:
        logging.debug("Creating build artifacts directory...")
        pathlib.Path(build_dir).mkdir(parents=True, exist_ok=True)
    image_core_path = os.path.join(build_dir, "rfnoc_image_core.sv")
    image_core_header_path = os.path.join(build_dir, "rfnoc_image_core.vh")
    secure_core_path = os.path.join(build_dir, "secure_image_core.sv")
    dts_path = os.path.join(build_dir, "device_tree.dts")
    makefile_srcs_path = os.path.join(build_dir, "Makefile.inc")
    fpga_top_dir = os.path.join(
        yaml_utils.get_top_path(os.path.abspath(repo_fpga_path)), target_dir(device)
    )

    # If desired, dump the configuration into the build directory
    if args.get("dump_config"):
        yaml_utils.write_yaml(config, os.path.join(build_dir, f"{image_core_name}.yml"))

    device_conf = yaml_utils.IOConfig(
        yaml_utils.device_config(core_config_path, device), signatures_conf
    )
    if not hasattr(device_conf, "top_dir"):
        setattr(device_conf, "top_dir", fpga_top_dir)
    for module_type, defs in known_modules.items():
        require_schema = None if module_type == "includes" else "rfnoc_modtool_args"
        yaml_utils.resolve_io_signatures(defs, signatures_conf, require_schema)

    # Load the image core config
    try:
        builder_conf = ImageBuilderConfig(config, known_modules, device_conf, include_paths)
    except (ValueError, KeyError) as e:
        logging.error("Error parsing image configuration: %s", e)
        return 1
    if not args.get("continue_on_warnings"):
        if builder_conf.warnings or builder_conf.errors:
            logging.error(
                "Image configuration contains issues: Skipping build. "
                "Use --ignore-warnings to build despite warnings."
            )
            return 1
    # Generate verilog output
    output_list = (
        [
            ("rfnoc_image_core.sv.mako", image_core_path, builder_conf),
            ("rfnoc_image_core.vh.mako", image_core_header_path, builder_conf),
            ("Makefile.inc.mako", makefile_srcs_path, builder_conf),
        ]
        + (
            [
                ("secure_image_core.sv.mako", secure_core_path, builder_conf.secure_config),
            ]
            if hasattr(builder_conf, "secure_image_core")
            else []
        )
        + (
            [
                ("device_tree.dts.mako", dts_path, builder_conf),
            ]
            if getattr(builder_conf.device, "generate_dts", False)
            else []
        )
    )
    reuse = args.get("reuse", False)
    for tpl, path, conf in output_list:
        if not (reuse and pathlib.Path(path).exists()):
            write_verilog(conf, path, args, template=tpl)
        else:
            logging.info(
                "Skipping generation of %s: File already exists and " "reuse was requested.", path
            )
    # Are we generating the secure image core?
    netlist_files = config.get("secure_image_core", {}).get("netlist_files")
    if args.get("secure_core"):
        secure_core_def = builder_conf.get_secure_core_def()
        secure_core_yml = args.get("secure_core")
        new_netlist_files = [
            "secure_image_core.vp",
            "secure_image_core.xdc",
        ]
        secure_core_def["secure_image_core"]["netlist_files"] = new_netlist_files
        logging.info("Writing secure core YAML to %s", secure_core_yml)
        yaml_utils.write_yaml(secure_core_def, secure_core_yml)
        if not args.get("secure_key_file"):
            viv_path = get_vivado_path(fpga_top_dir, args)
            key_dir = os.path.join(viv_path, "data", "pubkey")
            keyfiles = sorted(glob.glob(os.path.join(key_dir, "xilinxt*active.v")))
            if not keyfiles:
                logging.error("No key file found in %s", key_dir)
                return 1
            logging.info("No key file provided: Using default Vivado key from %s.", keyfiles[-1])
            with open(keyfiles[-1], "r") as f:
                write_verilog(
                    {},
                    os.path.join(build_dir, "keyfile.v"),
                    {"key_info": f.read()},
                    "keyfile.v.mako",
                )
            args["secure_key_file"] = os.path.join(build_dir, "keyfile.v")
    elif netlist_files:
        # The YAML includes a set of secure image netlist files. Copy them to
        # the artifact directory.
        netlist_dir = os.path.dirname(args.get("yaml_path"))
        for file in netlist_files:
            try:
                # Assume netlist file path is relative to YAML file
                source = os.path.join(netlist_dir, file)
                destination = os.path.normpath(os.path.join(build_dir, os.path.basename(file)))
                logging.info(f"Copying {file} to {destination}")
                shutil.copy(source, destination)
            except:
                logging.error(f"Failed to copy {source} to artifact directory")
                return 1

    ret_val = build(fpga_top_dir, device, build_dir, bool(netlist_files), base_dir, **args)
    if args.get("generate_only"):
        return ret_val

    if ret_val == 0 and args.get("secure_core"):
        # We built the secure image core netlist. Copy the resulting netlist
        # and constraint files.
        secure_core_yml_dir = os.path.dirname(args.get("secure_core"))
        file_list = [os.path.join(build_dir, x) for x in new_netlist_files]
        for file in file_list:
            try:
                logging.info(f"Copying {file} to {secure_core_yml_dir}")
                shutil.copy(file, secure_core_yml_dir)
            except:
                logging.warning(f"Unable to copy netlist file {file}")

    return ret_val



"""Image builder: Common definitions.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later
"""

# Supported protocol version
RFNOC_PROTO_VERSION = "1.0"

DEVICE_NAME = "_device_"

NONE_PORT = "_none_"



"""Image builder: Utilities and helpers.

Copyright 2019 Ettus Research, A National Instrument Brand

SPDX-License-Identifier: GPL-3.0-or-later

This module contains utility functions and helpers used by the image builder.
"""

import ast
import logging
import os
import re

from mako.exceptions import MakoException

from .template import Template


def merge_dicts(origd, newd):
    """Merge two dictionaries recursively.

    Originally from (with some modifications):
    https://stackoverflow.com/questions/71396284/python-how-to-recursively-merge-2-dictionaries
    """
    for key, val in newd.items():
        if key not in origd:
            origd[key] = val
            continue
        if isinstance(val, dict):
            if not isinstance(origd[key], dict):
                origd[key] = {}
            merge_dicts(origd[key], val)
        elif isinstance(val, (tuple, list)) and isinstance(origd[key], (tuple, list)):
            origd[key] = list(origd[key]) + list(val)
        else:
            origd[key] = val
    return origd


def generate_edge_table(config):
    """Generate the edge table based on the connections in 'config'.

    The edge table is a list of 4-tuples of the following format:

    (src_blk_index, src_port, dst_blk_index, dst_port)

    The block indices match the control crossbar indices. This table can be
    used to directly initialize the edge list in the FPGA.
    """
    block_con = [con for con in config.connections if con["srctype"] == "output"]
    logging.debug("Generating edge table...")
    edge_tbl = []
    for connection in block_con:
        if connection["srcblk"] in config.stream_endpoints:
            sep = config.stream_endpoints[connection["srcblk"]]
            index_match = re.match(r"out(\d)", connection["srcport"])
            if not index_match:
                logging.error(
                    "Port %s is invalid on endpoint %s", connection["srcport"], connection["srcblk"]
                )
            port_index = int(index_match.group(1))
            # Verify index < num_data_o
            if port_index >= sep["num_data_o"]:
                logging.error(
                    "Port %s exceeds num_data_o for endpoint %s",
                    connection["srcport"],
                    connection["srcblk"],
                )
            src = (sep["index"], port_index)
        else:
            block = config.noc_blocks[connection["srcblk"]]
            src = (block["index"], block["data"]["outputs"][connection["srcport"]]["index"])
        if connection["dstblk"] in config.stream_endpoints:
            sep = config.stream_endpoints[connection["dstblk"]]
            index_match = re.match(r"in(\d)", connection["dstport"])
            if not index_match:
                logging.error(
                    "Port %s is invalid on endpoint %s", connection["dstport"], connection["dstblk"]
                )
            # Verify index < num_data_i
            port_index = int(index_match.group(1))
            if port_index >= sep["num_data_i"]:
                logging.error(
                    "Port %s exceeds num_data_i for endpoint %s",
                    connection["dstport"],
                    connection["dstblk"],
                )
            dst = (sep["index"], port_index)
        else:
            block = config.noc_blocks[connection["dstblk"]]
            dst = (block["index"], block["data"]["inputs"][connection["dstport"]]["index"])
        edge_tbl.append((*src, *dst))
        logging.debug(
            "  %s-%s (%d,%d) => %s-%s (%d,%d)",
            connection["srcblk"],
            connection["srcport"],
            src[0],
            src[1],
            connection["dstblk"],
            connection["dstport"],
            dst[0],
            dst[1],
        )
    return edge_tbl


def resolve(var_val, **kwargs):
    """Resolve a string that contains variable references."""
    if not isinstance(var_val, str):
        return var_val
    if "env" not in kwargs:
        kwargs["env"] = dict(os.environ)
    try:
        tpl = Template(var_val)
        res = tpl.render(**kwargs)
    except MakoException as ex:
        raise SyntaxError(f"Unable to parse:\n{var_val}") from ex
    try:
        return ast.literal_eval(res)
    except (SyntaxError, ValueError):
        return res


def find_include_file(filename, include_paths):
    """Return an absolute path to a file based on a relative path.

    Factors in a list of include paths.

    When calling this, 'filename' contains a relative path to a file (e.g.,
    "rfnoc_block_foo/Makefile.srcs"). 'include_paths' is a list of paths in
    which this file is searched. Example:

    >>> find_include_file("rfnoc_block_foo/Makefile.srcs",
                          ["/usr/share/foo", "/usr/share/bar"])
    '/usr/share/foo/rfnoc_block_foo/Makefile.srcs'

    To determine the absolute path, it will search for 'filename' in all include
    paths in the given order.

    There are two exceptions, where the filename is returned verbatim:
    1. If the filename contains a make variable, then we just keep that and let
       make find the file, e.g. '$(LIB_DIR)/rfnoc/blocks/rfnoc_block_ddc/Makefile.srcs'
    2. If the filename is an absolute path, it is returned verbatim, but only
       if the file exists.

    If the file cannot be found, a FileNotFoundError is thrown.
    """
    # check for Makefile vars
    if re.match(r"\$\(.*\)", filename):
        return filename

    # if the file is already a fully valid path return the real path
    if os.path.isabs(filename):
        if os.path.isfile(filename):
            return os.path.realpath(filename)
        raise FileNotFoundError(f"File {filename} not found!")

    # otherwise search any combination of include_paths and filename
    for path in include_paths:
        candidate = os.path.join(path, filename)
        if os.path.isfile(candidate):
            return os.path.realpath(candidate)

    raise FileNotFoundError(f"File {filename} not found in include paths: {include_paths}")


def check_include_paths_backward_compat(include_paths):
    """Check if the include paths are plausible or maybe following an old style.

    In an older version of the image builder, the include paths for OOT builds
    pointed to the OOT modules itself (e.g., `-I ./rfnoc-gain`). However, the
    correct path should be the `rfnoc` subdir (e.g., `-I ./rfnoc-gain/rfnoc`).

    To make people's lives not unnecessarily hard, we allow the old behavior.
    This requires a heuristic to figure out if the user actually meant to use
    include_path + 'rfnoc' or not. This is done by checking if the include path
    contains expected subdirs (e.g., `blocks`, `fpga`) or if the path contains
    subdirs that are not expected (e.g., `cmake`, `rfnoc`).
    """

    def path_is_probably_old_style(path):
        """Return True if a path is missing an 'rfnoc' subdir."""
        expected_subdirs_new_style = ["blocks", "fpga", "transport_adapters", "modules"]
        expected_subdirs_old_style = ["cmake", "rfnoc"]
        if any(os.path.isdir(os.path.join(path, p)) for p in expected_subdirs_new_style):
            return False
        if all(os.path.isdir(os.path.join(path, p)) for p in expected_subdirs_old_style):
            return True
        return False

    def check_and_fix(path):
        if path_is_probably_old_style(path):
            new_path = os.path.join(path, "rfnoc")
            logging.info(
                "Include path %s is likely missing an 'rfnoc' subdir. Changing to: %s",
                path,
                new_path,
            )
            return os.path.join(path, "rfnoc")
        return path

    return [check_and_fix(path) for path in include_paths]
